"""
BufferedCanvas -- flicker-free canvas widget
Copyright (C) 2005, 2006 Daniel Keep, 2011 Duane Johnson

To use this widget, just override or replace the draw method.
This will be called whenever the widget size changes, or when
the update method is explicitly called.

Please submit any improvements/bugfixes/ideas to the following
url:

  http://wiki.wxpython.org/index.cgi/BufferedCanvas

2006-04-29: Added bugfix for a crash on Mac provided by Marc Jans.
"""

# Hint: try removing '.sp4msux0rz'
__author__ = 'Daniel Keep <daniel.keep.sp4msux0rz@gmail.com>'

__license__ = """
This file is part of the Printrun suite.

Printrun is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Printrun is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Printrun.  If not, see <http://www.gnu.org/licenses/>.
"""

__all__ = ['BufferedCanvas']

import wx

class BufferedCanvas(wx.Panel):
    """
    Implements a flicker-free canvas widget.

    Standard usage is to subclass this class, and override the
    draw method.  The draw method is passed a device context, which
    should be used to do your drawing.

    If you want to force a redraw (for whatever reason), you should
    call the update method.  This is because the draw method is never
    called as a result of an EVT_PAINT event.
    """

    # These are our two buffers.  Just be aware that when the buffers
    # are flipped, the REFERENCES are swapped.  So I wouldn't want to
    # try holding onto explicit references to one or the other ;)
    buffer = None
    backbuffer = None

    def __init__(self,
                 parent,
                 ID=-1,
                 pos = wx.DefaultPosition,
                 size = wx.DefaultSize,
                 style = wx.NO_FULL_REPAINT_ON_RESIZE | wx.WANTS_CHARS):
        wx.Panel.__init__(self, parent, ID, pos, size, style)

        # Bind events
        self.Bind(wx.EVT_PAINT, self.onPaint)

        # Disable background erasing (flicker-licious)
        def disable_event(*pargs, **kwargs):
            pass  # the sauce, please
        self.Bind(wx.EVT_ERASE_BACKGROUND, disable_event)

    #
    # General methods
    #

    def draw(self, dc, w, h):
        """
        Stub: called when the canvas needs to be re-drawn.
        """
        pass

    def update(self):
        """
        Causes the canvas to be updated.
        """
        self.Refresh()

    def getWidthHeight(self):
        width, height = self.GetClientSize()
        if width == 0:
            width = 1
        if height == 0:
            height = 1
        return (width, height)

    #
    # Event handlers
    #

    def onPaint(self, event):
        # Blit the front buffer to the screen
        w, h = self.GetClientSize()
        if not w or not h:
            return
        else:
            dc = wx.BufferedPaintDC(self)
            self.draw(dc, w, h)



# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017 Rock Storm <rockstorm@gmx.com>

import wx
from printrun.gui.widgets import get_space
from printrun.utils import install_locale
install_locale('pronterface')
# Set up Internationalization using gettext

class SpoolManagerMainWindow(wx.Frame):
    """
    Front-end for the Spool Manager.

    Main window which displays the currently loaded spools and the list of
    recorded ones with buttons to add, load, edit or delete them.
    """

    def __init__(self, parent, spool_manager):
        wx.Frame.__init__(self, parent,
                          title = _("Spool Manager"),
                          style = wx.DEFAULT_FRAME_STYLE)

        # An empty wx.Frame has a darker background on win, but filled with a panel it looks native
        self.panel = wx.Panel(self, -1)

        self.SetIcon(parent.GetIcon())

        # Initiate the back-end
        self.spool_manager = spool_manager
        self.spool_manager.refresh()

        # Generate the dialogs showing the current spools
        self.current_spools_dialog = CurrentSpoolDialog(self.panel,
                                                        self.spool_manager)

        # Check if any spools are loaded on non-exisiting extruders
        for spool in self.spool_manager.getSpoolList():
            if self.spool_manager.isLoaded(spool[0]) > (spool_manager.getExtruderCount() - 1):
                spool_manager.unload(self.spool_manager.isLoaded(spool[0]))

        # Generate the list of recorded spools
        self.spool_list = SpoolListView(self.panel, self.spool_manager)

        # Generate the buttons
        self.panel.new_button = wx.Button(self.panel, wx.ID_ADD)
        self.panel.new_button.SetToolTip(_("Add a new spool"))
        self.panel.edit_button = wx.Button(self.panel, wx.ID_EDIT)
        self.panel.edit_button.SetToolTip(_("Edit the selected spool"))
        self.panel.edit_button.Disable()
        self.panel.delete_button = wx.Button(self.panel, wx.ID_DELETE)
        self.panel.delete_button.SetToolTip(_("Delete the selected spool"))
        self.panel.delete_button.Disable()

        # Instead of a real statusbar, a virtual statusbar is combined with the close button
        self.statusbar = wx.StaticText(self.panel, -1, "", style = wx.ST_ELLIPSIZE_END)
        statusfont = wx.Font(self.statusbar.GetFont())
        statusfont.MakeSmaller()
        self.statusbar.SetFont(statusfont)

        self.close_button = wx.Button(self.panel, wx.ID_CLOSE)
        self.bottom_button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.bottom_button_sizer.Add(self.statusbar, 1, wx.ALIGN_CENTER_VERTICAL)
        self.bottom_button_sizer.Add(self.close_button, 0, wx.ALIGN_CENTER_VERTICAL)

        # "Program" the buttons
        self.panel.new_button.Bind(wx.EVT_BUTTON, self.onClickAdd)
        self.panel.edit_button.Bind(wx.EVT_BUTTON, self.onClickEdit)
        self.panel.delete_button.Bind(wx.EVT_BUTTON, self.onClickDelete)

        self.close_button.Bind(wx.EVT_BUTTON, self.onClickClose)

        # Layout
        ## Group the buttons
        self.list_button_sizer = wx.BoxSizer(wx.VERTICAL)
        self.list_button_sizer.Add(self.panel.new_button, 0,
            wx.FIXED_MINSIZE | wx.EXPAND | wx.LEFT | wx.BOTTOM, get_space('minor'))
        self.list_button_sizer.Add(self.panel.edit_button, 0,
            wx.FIXED_MINSIZE | wx.EXPAND | wx.LEFT | wx.BOTTOM, get_space('minor'))
        self.list_button_sizer.Add(self.panel.delete_button, 0,
            wx.FIXED_MINSIZE | wx.EXPAND | wx.LEFT, get_space('minor'))

        ## Group the buttons with the spool list
        self.list_sizer = wx.StaticBoxSizer(wx.HORIZONTAL, self.panel, label = _("Spool List"))
        self.list_sizer.Add(self.spool_list, 1,
                            wx.EXPAND | wx.LEFT | wx.TOP | wx.BOTTOM, get_space('staticbox'))
        self.list_sizer.Add(self.list_button_sizer, 0,
                            wx.ALIGN_TOP | wx.TOP | wx.RIGHT, get_space('staticbox'))

        ## Layout the whole thing
        widgetsizer = wx.BoxSizer(wx.VERTICAL)
        widgetsizer.Add(self.list_sizer, 1, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, get_space('minor'))
        widgetsizer.Add(self.current_spools_dialog, 0,
                        wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM, get_space('minor'))
        widgetsizer.Add(wx.StaticLine(self.panel, -1, style = wx.LI_HORIZONTAL), 0, wx.EXPAND)
        widgetsizer.Add(self.bottom_button_sizer, 0, wx.EXPAND | wx.ALL, get_space('stddlg-frame'))

        ## Make sure the frame has the right size when it opens, but can still be resized
        self.panel.SetSizer(widgetsizer)
        topsizer = wx.BoxSizer(wx.VERTICAL)
        topsizer.Add(self.panel, -1, wx.EXPAND)
        self.SetSizer(topsizer)
        self.SetMinClientSize(self.panel.GetEffectiveMinSize())
        self.Fit()
        self.CentreOnParent()

    def onClickAdd(self, event):
        """Open the window for customizing the new spool."""
        SpoolManagerAddWindow(self).ShowModal()

    def onClickLoad(self, event, extruder):
        """Load the selected spool to the correspondent extruder."""

        # Check whether there is a spool selected
        spool_index = self.spool_list.GetFirstSelected()
        if spool_index == -1:
            self.statusbar.SetLabel(
                _("Could not load the spool. No spool selected."))
            return
        spool_name = self.spool_list.GetItemText(spool_index)
        self.statusbar.SetLabel("")

        # If selected spool is already loaded, do nothing
        spool_extruder = self.spool_manager.isLoaded(spool_name)
        if spool_extruder > -1:
            self.statusbar.SetLabel(
                _("Spool '%s' is already loaded for Extruder %d.") %
                (spool_name, spool_extruder))
            self.Layout()  # Layout() is needed to ellipsize possible overlength status
            return

        # Load the selected spool and refresh the current spools dialog
        self.spool_manager.load(spool_name, extruder)
        self.current_spools_dialog.refreshDialog(self.spool_manager)
        self.current_spools_dialog.unload_button[extruder].Enable()
        self.statusbar.SetLabel(
            _("Loaded spool '%s' for Extruder %d.") % (spool_name, extruder))
        self.Layout()  # Layout() is needed to ellipsize possible overlength status

    def onClickUnload(self, event, extruder):
        """Unload the spool from the correspondent extruder."""

        spool_name = self.spool_manager.getSpoolName(extruder)
        if spool_name is not None:
            self.spool_manager.unload(extruder)
            self.current_spools_dialog.refreshDialog(self.spool_manager)
            self.statusbar.SetLabel(
                _("Unloaded spool from Extruder %d.") % extruder)
            self.current_spools_dialog.unload_button[extruder].Disable()
        else:
            self.statusbar.SetLabel(
                _("There is no spool loaded for Extruder %d.") % extruder)

    def onClickEdit(self, event):
        """Open the window for editing the data of the selected spool."""

        # Check whether there is a spool selected
        spool_index = self.spool_list.GetFirstSelected()
        if spool_index == -1:
            self.statusbar.SetLabel(
                _("Could not edit the spool. No spool selected."))
            return

        # Open the edit window
        spool_name = self.spool_list.GetItemText(spool_index)
        spool_length = self.spool_list.GetItemText(spool_index, 1)
        SpoolManagerEditWindow(self, spool_name, spool_length).ShowModal()
        self.statusbar.SetLabel("")

    def onClickDelete(self, event):
        """Delete the selected spool."""

        # Get the selected spool
        spool_index = self.spool_list.GetFirstSelected()
        if spool_index == -1:
            self.statusbar.SetLabel(
                _("Could not delete the spool. No spool selected."))
            return
        spool_name = self.spool_list.GetItemText(spool_index)
        self.statusbar.SetLabel("")

        # Ask confirmation for deleting
        delete_dialog = wx.MessageDialog(self,
                                         message = _("Are you sure you want to delete the '%s' spool?") %
                                         spool_name, caption = _("Delete Spool"),
                                         style = wx.YES_NO | wx.ICON_EXCLAMATION)

        if delete_dialog.ShowModal() == wx.ID_YES:
            # Remove spool
            self.spool_manager.remove(spool_name)
            self.spool_list.refreshList(self.spool_manager)
            self.current_spools_dialog.refreshDialog(self.spool_manager)
            self.statusbar.SetLabel(
                _("Deleted spool '%s'.") % spool_name)

    def onClickClose(self, event):
        self.Destroy()

class SpoolListView(wx.ListView):
    """
    Custom wxListView object which visualizes the list of available spools.
    """

    def __init__(self, parent, spool_manager):
        wx.ListView.__init__(self, parent,
                             style = wx.LC_REPORT | wx.LC_SINGLE_SEL)
        self.InsertColumn(0, _("Spool"), width = wx.LIST_AUTOSIZE_USEHEADER)
        self.InsertColumn(1, _("Filament"), width = wx.LIST_AUTOSIZE_USEHEADER)
        self.populateList(spool_manager)

        # "Program" the layout
        self.Bind(wx.EVT_SIZE, self.onResizeList)
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.onItemSelect)
        self.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.onItemDeselect)
        self.Bind(wx.EVT_LIST_DELETE_ITEM, self.onItemDeselect)
        self.Bind(wx.EVT_LIST_INSERT_ITEM, self.onItemDeselect)

    def populateList(self, spool_manager):
        """Get the list of recorded spools from the Spool Manager."""
        spool_list = spool_manager.getSpoolList()
        for spool in spool_list:
            spool[1] = str(spool[1]) + " mm"
            self.Append(spool)

    def refreshList(self, spool_manager):
        """Refresh the list by re-reading the Spool Manager list."""
        self.DeleteAllItems()
        self.populateList(spool_manager)

    def onResizeList(self, event):
        list_size = self.GetSize()
        self.SetColumnWidth(1, -2)
        filament_column_width = self.GetColumnWidth(1)
        self.SetColumnWidth(col = 0,
                            width = list_size.width - filament_column_width)
        event.Skip()

    def onItemSelect(self, event):
        self.Parent.edit_button.Enable()
        self.Parent.delete_button.Enable()

    def onItemDeselect(self, event):
        self.Parent.edit_button.Disable()
        self.Parent.delete_button.Disable()

class CurrentSpoolDialog(wx.Panel):
    """
    Custom wxStaticText object to display the currently loaded spools and
    their remaining filament.
    """

    def __init__(self, parent, spool_manager):
        wx.Panel.__init__(self, parent)
        self.parent = parent
        self.extruders = spool_manager.getExtruderCount()
        # If the settings file has no entry, at least 1 extruder will be set
        if not self.extruders:
            self.extruders = 1

        csd_sizer = wx.BoxSizer(wx.VERTICAL)

        # Calculate the minimum size needed to properly display the
        # extruder information
        min_size = self.GetTextExtent("Default Very Long Spool Name")

        # Generate a dialog for every extruder
        self.extruder_dialog = []
        load_button = []
        self.unload_button = []
        button_sizer = []
        dialog_sizer = []
        for i in range(self.extruders):
            # Generate the dialog with the spool information
            textlabel = wx.StaticText(self, label = _("Name:\nRemaining filament:"),
                                      style = wx.ALIGN_RIGHT)
            self.extruder_dialog.append(
                wx.StaticText(self, style = wx.ST_ELLIPSIZE_END))
            self.extruder_dialog[i].SetMinSize(wx.Size(min_size.width, -1))

            # Generate the "load" and "unload" buttons
            load_button.append(wx.Button(self, label = _("Load")))
            load_button[i].SetToolTip(
                _("Load selected spool for Extruder %d") % i)
            self.unload_button.append(wx.Button(self, label = _("Unload")))
            self.unload_button[i].Disable()
            self.unload_button[i].SetToolTip(
                _("Unload the spool for Extruder %d") % i)

            # "Program" the buttons
            load_button[i].Bind(wx.EVT_BUTTON,
                lambda event, extruder=i: parent.Parent.onClickLoad(event, extruder))
            self.unload_button[i].Bind(wx.EVT_BUTTON,
                lambda event, extruder=i: parent.Parent.onClickUnload(event, extruder))

            # Layout
            button_sizer.append(wx.BoxSizer(wx.HORIZONTAL))
            button_sizer[i].Add(load_button[i], 0,
                wx.FIXED_MINSIZE | wx.ALIGN_CENTER | wx.RIGHT, get_space('minor'))
            button_sizer[i].Add(self.unload_button[i], 0,
                wx.FIXED_MINSIZE | wx.ALIGN_CENTER)

            dialog_sizer.append(wx.StaticBoxSizer(wx.HORIZONTAL,
                                                  self, label = _("Spool for Extruder %d:") % i))
            dialog_sizer[i].Add(textlabel, 0, wx.ALIGN_TOP | wx.ALL, get_space('staticbox'))
            dialog_sizer[i].AddSpacer(get_space('minor'))
            dialog_sizer[i].Add(self.extruder_dialog[i], 1, wx.ALIGN_TOP | wx.TOP, get_space('staticbox'))
            dialog_sizer[i].AddSpacer(get_space('major'))
            dialog_sizer[i].Add(button_sizer[i], 0, wx.EXPAND | wx.RIGHT, get_space('staticbox'))

            csd_sizer.Add(dialog_sizer[i], 0, wx.EXPAND | wx.TOP, get_space('minor'))

        self.refreshDialog(spool_manager)

        self.SetSizerAndFit(csd_sizer)

    def refreshDialog(self, spool_manager):
        """Retrieve the current spools from the Spool Manager."""

        for i in range(self.extruders):
            spool_name = spool_manager.getSpoolName(i)
            if spool_name is not None:
                self.unload_button[i].Enable()
            else:
                self.unload_button[i].Disable()
            spool_filament = spool_manager.getRemainingFilament(i)
            label = ("%s\n" % spool_name +
                     "%.2f mm" % spool_filament)
            self.extruder_dialog[i].SetLabelText(label)


# ---------------------------------------------------------------------------
def checkOverwrite(parent, spool_name):
    """Ask the user whether or not to overwrite the existing spool."""

    overwrite_dialog = wx.MessageDialog(parent,
        message = _("A spool with the name '%s'' already exists.") %
            spool_name +
            _("Do you wish to overwrite it?"),
            caption = _("Overwrite"),
            style = wx.YES_NO | wx.ICON_EXCLAMATION)

    return overwrite_dialog.ShowModal() == wx.ID_YES


def getFloat(parent, number):
    """
    Check whether the input number is a float. Either return the number or
    return False.
    """
    if ',' in number:
        parent.parent.statusbar.SetLabel(_("Value contains a comma, please use a point for decimal values: %s") % number)
        parent.parent.Layout()  # Layout() is needed to ellipsize possible overlength status
        return False

    try:
        return float(number)
    except ValueError:
        parent.parent.statusbar.SetLabel(_("Unrecognized number: %s") % number)
        parent.parent.Layout()  # Layout() is needed to ellipsize possible overlength status
        return False

# ---------------------------------------------------------------------------
class SpoolManagerAddWindow(wx.Dialog):
    """Window for adding spools."""

    def __init__(self, parent):

        wx.Dialog.__init__(self, parent,
                           title = _("Add Spool"),
                           style = wx.DEFAULT_DIALOG_STYLE)

        self.parent = parent

        self.SetIcon(parent.GetIcon())

        # Generate the dialogs
        # The wx.TextCtrl variabels need to be declared before the loop, empty
        self.name_dialog = wx.TextCtrl(self, -1)
        self.diameter_dialog = wx.TextCtrl(self, -1)
        self.weight_dialog = wx.TextCtrl(self, -1)
        self.density_dialog = wx.TextCtrl(self, -1)
        self.length_dialog = wx.TextCtrl(self, -1)

        # The list contains field-description, textctrl value, default value, unit, tooltip;
        name_dlg = [_("Name:"), self.name_dialog, _("Default Spool"), "", ""]
        diameter_dlg = [_("Diameter:"), self.diameter_dialog, "1.75", "mm",
                        _("Typically, either 1.75 mm or 2.85 mm")]
        weight_dlg = [_("Weight:"), self.weight_dialog, "1.0", "kg", ""]
        density_dlg = [_("Density:"), self.density_dialog, "1.25", "g/cm^3",
                       _("Typical densities are 1.25 g/cm^3 for PLA,\n1.27 g/cm^3 for PETG or 1.08 g/cm^3 for ABS")]
        length_dlg = [_("Length:"), self.length_dialog, "332601.35", "mm", ""]
        dialog_list = [name_dlg, diameter_dlg, weight_dlg, density_dlg, length_dlg]

        minwidth = self.GetTextExtent('Default Long Spool Name').width
        grid = wx.FlexGridSizer(rows = 0, cols = 3, hgap = get_space('minor'), vgap = get_space('minor'))

        for dialog in dialog_list:
            # Add a field-description label
            grid.Add(wx.StaticText(self, -1, dialog[0], size = (-1, -1)), 0,
                     wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL)
            # Give the TextCtrl the right value
            dialog[1].ChangeValue(dialog[2])
            dialog[1].SetMinSize((minwidth, -1))
            # Add a tooltip
            if dialog[4] != "":
                dialog[1].SetToolTip(dialog[4])
            grid.Add(dialog[1])
            # Add a label for the unit
            if dialog[3] == "":
                grid.Add((0, 0))
            else:
                grid.Add(wx.StaticText(self, -1, dialog[3], size = (-1, -1)), 0, wx.ALIGN_CENTER_VERTICAL)

        # "Program" the dialogs
        self.diameter_dialog.Bind(wx.EVT_TEXT, self.calculateLength)
        self.weight_dialog.Bind(wx.EVT_TEXT, self.calculateLength)
        self.density_dialog.Bind(wx.EVT_TEXT, self.calculateLength)
        self.length_dialog.Bind(wx.EVT_TEXT, self.calculateWeight)

        # Generate the bottom buttons
        self.add_button = wx.Button(self, wx.ID_ADD)
        self.cancel_button = wx.Button(self, wx.ID_CANCEL)

        # "Program" the bottom buttons
        self.add_button.Bind(wx.EVT_BUTTON, self.onClickAdd)
        self.add_button.SetDefault()
        self.SetAffirmativeId(wx.ID_ADD)
        self.cancel_button.Bind(wx.EVT_BUTTON, self.onClickCancel)

        # Layout
        ## Setup the bottom buttons
        self.bottom_buttons_sizer = wx.StdDialogButtonSizer()
        self.bottom_buttons_sizer.SetAffirmativeButton(self.add_button)
        self.bottom_buttons_sizer.AddButton(self.cancel_button)

        self.bottom_buttons_sizer.Realize()

        ## Group the whole window
        self.topsizer = wx.BoxSizer(wx.VERTICAL)
        self.topsizer.Add(grid, 1, wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT, get_space('major'))
        self.topsizer.Add(wx.StaticLine(self, -1, style = wx.LI_HORIZONTAL), 0,
                          wx.EXPAND | wx.TOP, get_space('minor'))
        self.topsizer.Add(self.bottom_buttons_sizer, 0,
                          wx.ALIGN_RIGHT | wx.ALL, get_space('stddlg'))

        self.SetSizerAndFit(self.topsizer)
        self.CentreOnParent()
        self.name_dialog.SetFocus()

    def onClickAdd(self, ev):
        """Add the new spool and close the window."""

        spool_name = self.name_dialog.GetValue()
        spool_length = getFloat(self, self.length_dialog.GetValue())

        # Check whether the length is actually a number
        if not spool_length:
            self.parent.statusbar.SetLabel(_("ERROR: Unrecognized length: %s.") %
                                           self.length_dialog.GetValue())
            self.parent.Layout()  # Layout() is needed to ellipsize possible overlength status
            return

        # The remaining filament should always be a positive number
        if not spool_length > 0:
            self.parent.statusbar.SetLabel(_("ERROR: Length is zero or negative: %.2f.") %
                                           spool_length)
            self.parent.Layout()  # Layout() is needed to ellipsize possible overlength status
            return

        # Check whether the name is already used. If it is used, prompt the
        # user before overwriting it
        if self.parent.spool_manager.isListed(spool_name):
            if checkOverwrite(self, spool_name):
                # Remove the "will be overwritten" spool
                self.parent.spool_manager.remove(spool_name)
            else:
                return

        # Add the new spool
        self.parent.spool_manager.add(spool_name, spool_length)
        self.parent.spool_list.refreshList(self.parent.spool_manager)
        self.parent.current_spools_dialog.refreshDialog(
                self.parent.spool_manager)
        self.parent.statusbar.SetLabel(
            _("Added new spool '%s'") % spool_name +
            _(" with %.2f mm of remaining filament.") % spool_length)
        self.parent.Layout()  # Layout() is needed to ellipsize possible overlength status

        self.EndModal(True)
        self.Destroy()

    def onClickCancel(self, event):
        """Do nothing and close the window."""
        self.parent.statusbar.SetLabel("")
        self.EndModal(True)
        self.Destroy()

    def calculateLength(self, event):
        """
        Calculate the length of the filament given the mass, diameter and
        density of the filament. Set the 'Length' field to this quantity.
        """

        mass = getFloat(self, self.weight_dialog.GetValue())
        diameter = getFloat(self, self.diameter_dialog.GetValue())
        density = getFloat(self, self.density_dialog.GetValue())
        if mass and diameter and density:
            pi = 3.14159265359
            length = 4e6 * mass / pi / diameter**2 / density
            self.parent.statusbar.SetLabel("")
            self.length_dialog.ChangeValue("%.2f" % length)
        else:
            self.length_dialog.ChangeValue("---")

    def calculateWeight(self, event):
        """
        Calculate the weight of the filament given the length, diameter and
        density of the filament. Set the 'Weight' field to this value.
        """

        length = getFloat(self, self.length_dialog.GetValue())
        diameter = getFloat(self, self.diameter_dialog.GetValue())
        density = getFloat(self, self.density_dialog.GetValue())
        if length and diameter and density:
            pi = 3.14159265359
            mass = length * pi * diameter**2 * density / 4e6
            self.parent.statusbar.SetLabel("")
            self.weight_dialog.ChangeValue("%.2f" % mass)
        else:
            self.weight_dialog.ChangeValue("---")

# ---------------------------------------------------------------------------
class SpoolManagerEditWindow(wx.Dialog):
    """Window for editing the name or the length of a spool."""

    def __init__(self, parent, spool_name, spool_length):

        wx.Dialog.__init__(self, parent,
            title = _("Edit Spool"),
            style = wx.DEFAULT_DIALOG_STYLE)

        self.parent = parent

        self.SetIcon(parent.GetIcon())

        self.old_spool_name = spool_name
        self.old_spool_length = getFloat(self, spool_length.replace(" mm", ""))

        # Set how many millimeters will the buttons add or subtract
        self.quantities = [-100.0, -50.0, -10.0, 10.0, 50.0, 100.0]

        # Generate the name field
        self.name_title = wx.StaticText(self, -1, _("Name:"))
        minwidth = self.GetTextExtent('Default Very Long Spool Name').width
        self.name_field = wx.TextCtrl(self, -1, self.old_spool_name, style = wx.TE_RIGHT)
        self.name_field.SetMinSize((minwidth, -1))

        # Generate the length field
        self.length_title = wx.StaticText(self, label = _("Remaining filament:"),
                                          style = wx.ALIGN_RIGHT)
        self.length_field = wx.TextCtrl(self, -1, value = str(self.old_spool_length),
                                        style = wx.TE_RIGHT)
        self.length_field.SetMinSize((minwidth, -1))

        # Generate the buttons
        button_min_width = self.GetTextExtent('  +000.0  ').width
        self.minus3_button = wx.Button(self,
            label = str(self.quantities[0]))
        self.minus2_button = wx.Button(self,
            label = str(self.quantities[1]))
        self.minus1_button = wx.Button(self,
            label = str(self.quantities[2]))

        self.plus1_button = wx.Button(self,
            label = "+" + str(self.quantities[3]))
        self.plus2_button = wx.Button(self,
            label = "+" + str(self.quantities[4]))
        self.plus3_button = wx.Button(self,
            label = "+" + str(self.quantities[5]))

        self.minus3_button.SetSize((button_min_width, -1))
        self.minus2_button.SetSize((button_min_width, -1))
        self.minus1_button.SetSize((button_min_width, -1))
        self.plus1_button.SetSize((button_min_width, -1))
        self.plus2_button.SetSize((button_min_width, -1))
        self.plus3_button.SetSize((button_min_width, -1))

        # "Program" the length buttons
        self.minus3_button.Bind(wx.EVT_BUTTON, self.changeLength)
        self.minus2_button.Bind(wx.EVT_BUTTON, self.changeLength)
        self.minus1_button.Bind(wx.EVT_BUTTON, self.changeLength)
        self.plus1_button.Bind(wx.EVT_BUTTON, self.changeLength)
        self.plus2_button.Bind(wx.EVT_BUTTON, self.changeLength)
        self.plus3_button.Bind(wx.EVT_BUTTON, self.changeLength)

        # Generate the bottom buttons
        self.save_button = wx.Button(self, wx.ID_SAVE)
        self.cancel_button = wx.Button(self, wx.ID_CANCEL)

        # "Program" the bottom buttons
        self.save_button.Bind(wx.EVT_BUTTON, self.onClickSave)
        self.cancel_button.Bind(wx.EVT_BUTTON, self.onClickCancel)
        self.save_button.SetDefault()
        self.SetAffirmativeId(wx.ID_SAVE)

        # Layout
        ## Group the length field and its correspondent buttons
        self.btn_sizer = wx.StaticBoxSizer(wx.HORIZONTAL, self)
        self.btn_sizer.Add(self.minus3_button, 0,
                           wx.FIXED_MINSIZE | wx.ALIGN_CENTER | wx.LEFT | wx.TOP | wx.BOTTOM,
                           get_space('staticbox'))
        self.btn_sizer.Add(self.minus2_button, 0,
                           wx.FIXED_MINSIZE | wx.ALIGN_CENTER | wx.LEFT | wx.RIGHT, get_space('mini'))
        self.btn_sizer.Add(self.minus1_button, 0,
                           wx.FIXED_MINSIZE | wx.ALIGN_CENTER | wx.RIGHT, get_space('mini'))
        self.btn_sizer.AddSpacer(get_space('major'))
        self.btn_sizer.Add(self.plus1_button, 0,
                           wx.FIXED_MINSIZE | wx.ALIGN_CENTER | wx.LEFT, get_space('mini'))
        self.btn_sizer.Add(self.plus2_button, 0,
                           wx.FIXED_MINSIZE | wx.ALIGN_CENTER | wx.LEFT | wx.RIGHT, get_space('mini'))
        self.btn_sizer.Add(self.plus3_button, 0,
                           wx.FIXED_MINSIZE | wx.ALIGN_CENTER | wx.RIGHT, get_space('staticbox'))

        ## Group the bottom buttons
        self.bottom_buttons_sizer = wx.StdDialogButtonSizer()
        self.bottom_buttons_sizer.AddButton(self.save_button)
        self.bottom_buttons_sizer.AddButton(self.cancel_button)
        self.bottom_buttons_sizer.Realize()

        ## Lay out the whole window
        grid = wx.GridBagSizer(hgap = get_space('minor'), vgap = get_space('minor'))

        # Gridbagsizer: pos = (row, col), span = (rowspan, colspan)
        grid.Add(self.name_title, pos = (0, 0), span = (1, 1), border = 0,
                 flag = wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND)
        grid.Add(self.name_field, pos = (0, 1), span = (1, 1), border = 0,
                 flag = wx.ALIGN_LEFT)
        grid.Add(self.length_title, pos = (1, 0), span = (1, 1), border = 0,
                 flag = wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND)
        grid.Add(self.length_field, pos = (1, 1), span = (1, 1), border = 0,
                 flag = wx.ALIGN_LEFT)
        grid.Add(self.btn_sizer, pos = (2, 0), span = (1, 2), border = 0,
                 flag = wx.ALIGN_CENTER | wx.EXPAND)

        topsizer = wx.BoxSizer(wx.VERTICAL)
        topsizer.Add(grid, 1, wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT, get_space('major'))
        topsizer.Add(wx.StaticLine(self, -1, style = wx.LI_HORIZONTAL), 0,
                     wx.EXPAND | wx.TOP, get_space('minor'))
        topsizer.Add(self.bottom_buttons_sizer, 0, wx.ALIGN_RIGHT | wx.ALL, get_space('stddlg'))

        self.SetSizer(topsizer)
        self.Fit()
        self.CentreOnParent()
        self.name_field.SetFocus()

    def changeLength(self, event):
        new_length = getFloat(self, self.length_field.GetValue())
        if new_length:
            new_length = new_length + float(event.GetEventObject().GetLabel())
            self.length_field.ChangeValue("%.2f" % new_length)
            self.parent.statusbar.SetLabel("")

    def onClickSave(self, event):

        new_spool_name = self.name_field.GetValue()
        new_spool_length = getFloat(self, self.length_field.GetValue())

        # Check whether the length is actually a number
        if not new_spool_length:
            self.parent.statusbar.SetLabel(
                _("ERROR: Unrecognized length: %s.") %
                self.length_field.GetValue())
            self.parent.Layout()  # Layout() is needed to ellipsize possible overlength status
            return

        if not new_spool_length > 0:
            self.parent.statusbar.SetLabel(
                _("ERROR: Length is zero or negative: %.2f.") % new_spool_length)
            self.parent.Layout()  # Layout() is needed to ellipsize possible overlength status
            return

        # Check whether the "old" spool was loaded
        new_spool_extruder = self.parent.spool_manager.isLoaded(
            self.old_spool_name)

        # Check whether the name has changed
        if new_spool_name == self.old_spool_name:
            # Remove only the "old" spool
            self.parent.spool_manager.remove(self.old_spool_name)
        else:
            # Check whether the new name is already used
            if self.parent.spool_manager.isListed(new_spool_name):
                if checkOverwrite(self, new_spool_name):
                    # Remove the "old" and the "will be overwritten" spools
                    self.parent.spool_manager.remove(self.old_spool_name)
                    self.parent.spool_manager.remove(new_spool_name)
                else:
                    return
            else:
                # Remove only the "old" spool
                self.parent.spool_manager.remove(self.old_spool_name)

        # Add "new" or edited spool
        self.parent.spool_manager.add(new_spool_name, new_spool_length)
        self.parent.spool_manager.load(new_spool_name, new_spool_extruder)
        self.parent.spool_list.refreshList(self.parent.spool_manager)
        self.parent.current_spools_dialog.refreshDialog(
            self.parent.spool_manager)
        self.parent.statusbar.SetLabel(
            _("Edited spool '%s'") % new_spool_name +
            _(" with %.2f mm of remaining filament.") % new_spool_length)
        self.parent.Layout()  # Layout() is needed to ellipsize possible overlength status
        self.EndModal(True)
        self.Destroy()

    def onClickCancel(self, event):
        self.parent.statusbar.SetLabel("")
        self.EndModal(True)
        self.Destroy()



# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017 Rock Storm <rockstorm@gmx.com>

# This module indirectly depends of pronsole and settings but it does not
# import them

class SpoolManager():
    """
    Back-end for the Spool Manager.

    It is expected to be called from an object which has the contents of
    settings.py and pronsole.py. This way the class is able to '_add' and
    'set' settings.

    This class basically handles a single variable called '_spool_list'. It is
    a list of spool_items. A spool_item is in turn a list three elements: a
    string, a float and an integer. Namely: the name of the spool, the
    remaining length of filament and the extruder it is loaded to. E.g.:

       spool_item = [string name, float length, int extruder]

       _spool_list = [spool_item spool_1, ... , spool_item spool_n ]

    '_spool_list' is somehow a Nx3 matrix where N is the number of recorded
    spools. The first column contains the names of the spools, the second the
    lengths of remaining filament and the third column contains which extruder
    is the spool loaded for.

    The variable '_spool_list' is saved in the configuration file using a
    setting with the same name: 'spool_list'. It is saved as a single string.
    It concatenates every item from the list and separates them by a comma and
    a space. For instance, if the variable '_spool_list' was:

           _spool_list = [["spool_1", 100.0, 0], ["spool_2", 200.0, -1]]

       The 'spool_list' setting will look like:

           "spool_1, 100.0, 0, spool_2, 200.0, -1"
    """

    def __init__(self, parent):
        self.parent = parent
        self.refresh()

    def refresh(self):
        """
        Read the configuration file and populate the list of recorded spools.
        """
        self._spool_list = self._readSetting(self.parent.settings.spool_list)

    def add(self, spool_name, spool_length):
        """Add the given spool to the list of recorded spools."""
        self._spool_list.append([spool_name, spool_length, -1])
        self._save()

    def load(self, spool_name, extruder):
        """Set the extruder field of the given spool item."""

        # If there was a spool already loaded for this extruder unload it
        previous_spool = self._findByColumn(extruder, 2)
        if previous_spool != -1:
            self.unload(extruder)

        # Load the given spool
        new_spool = self._findByColumn(spool_name, 0)
        self.remove(spool_name)
        self._spool_list.append([new_spool[0], new_spool[1], extruder])
        self._save()

    def remove(self, spool_name):
        """Remove the given spool item from the list of recorded spools."""
        spool_item = self._findByColumn(spool_name, 0)
        self._spool_list.remove(spool_item)
        self._save()

    def unload(self, extruder):
        """Set to -1 the extruder field of the spool item currently on."""

        spool_item = self._findByColumn(extruder, 2)
        if spool_item != -1:
            self.remove(spool_item[0])
            self._spool_list.append([spool_item[0], spool_item[1], -1])
            self._save()

    def isLoaded(self, spool_name):
        """
        int isLoaded( string name )

        Return the extruder that the given spool is loaded to. -1 if it is
        not loaded for any extruder or None if the given name does not match
        any known spool.
        """

        spool_item = self._findByColumn(spool_name, 0)
        if spool_item != -1:
            return spool_item[2]
        else:
            return None

    def isListed(self, spool_name):
        """Return 'True' if the given spool is on the list."""

        spool_item = self._findByColumn(spool_name, 0)
        if not spool_item == -1:
            return True
        else:
            return False

    def getSpoolName(self, extruder):
        """
        string getSpoolName( int extruder )

        Return the name of the spool loaded for the given extruder.
        """

        spool_item = self._findByColumn(extruder, 2)
        if spool_item != -1:
            return spool_item[0]
        else:
            return None

    def getRemainingFilament(self, extruder):
        """
        float getRemainingFilament( int extruder )

        Return the name of the spool loaded for the given extruder.
        """

        spool_item = self._findByColumn(extruder, 2)
        if spool_item != -1:
            return spool_item[1]
        else:
            return float("NaN")

    def editLength(self, increment, spool_name = None, extruder = -1):
        """
        int editLength ( float increment, string spool_name, int extruder )

        Add the given 'increment' amount to the length of filament of the
        given spool. Spool can be specified either by name or by the extruder
        it is loaded to.
        """

        if spool_name != None:
            spool_item = self._findByColumn(spool_name, 0)
        elif extruder != -1:
            spool_item = self._findByColumn(extruder, 2)
        else:
            return -1   # Not enough arguments

        if spool_item == -1:
            return -2   # No spool found for the given name or extruder

        length = spool_item[1] + increment
        self.remove(spool_item[0])
        self.add(spool_item[0], length)
        if spool_item[2] > -1:
            self.load(spool_item[0], spool_item[2])
        self._save()

        return 0

    def getExtruderCount(self):
        """int getExtruderCount()"""
        return self.parent.settings.extruders

    def getSpoolCount(self):
        """
        int getSpoolCount()

        Return the number of currently recorded spools.
        """
        return len(self._spool_list)

    def getSpoolList(self):
        """
        [N][2] getSpoolList ()

        Returns a list of the recorded spools. Returns a Nx2 matrix where N is
        the number of recorded spools. The first column contains the names of
        the spools and the second the lengths of remaining filament.
        """

        slist = []
        for i in range(self.getSpoolCount()):
            item = [self._spool_list[i][0], self._spool_list[i][1]]
            slist.append(item)
        return slist

    def _findByColumn(self, data, col = 0):
        """
        Find which spool_item from the list contains certain data.

        The 'col' argument specifies in which field from the spool_item to
        look for. For instance, with the following list:

            _spool_list = [["spool_1",   100.0, 1],
                           ["spool_2",   200.0, 0],
                           .
                           .
                           .
                           ["spool_10", 1000.0, 0]]

        A call like: _findByColumn("spool_2", 0)

        Will produce: ["spool_2", 200.0, 0]

        col = 0, would look into the "name's column"
        col = 1, would look into the "length's column"
        col = 2, would look into the "extruder's column"
        """

        for spool_item in self._spool_list:
            if data == spool_item[col]:
                return spool_item

        return -1

    def _save(self):
        """Update the list of recorded spools in the configuration file."""
        self._setSetting(self._spool_list, "spool_list")

    def _setSetting(self, variable, setting):
        """
        Write the given variable to the given setting of the configuration
        file.
        """
        n = 3 # number of fields in spool_item
        string_list = []
        for i in range(len(variable)):
            for j in range(n):
                string_list.append(str(variable[i][j]))
        separator = ", "
        self.parent.set(setting, separator.join(string_list))

    def _readSetting(self, setting):
        """
        Return the variable read.
        """
        n = 3 # number of fields in spool_item
        string_list = setting.split(", ")
        variable = []
        for i in range(len(string_list)//n):
            variable.append(
                [string_list[n*i],
                 float(string_list[n*i+1]),
                 int(string_list[n*i+2])])
        return variable



#!/usr/bin/python

# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.

import sys
import traceback
import logging
from printrun.pronsole import pronsole
import getopt

if __name__ == "__main__":

    from printrun.printcore import __version__ as printcore_version

    usage = "Usage:\n"+\
            "  pronsole [OPTIONS] [FILE]\n\n"+\
            "Options:\n"+\
            "  -h, --help\t\t\tPrint this help message and exit\n"+\
            "  -V, --version\t\t\tPrint program's version number and exit\n"+\
            "  -v, --verbose\t\t\tIncrease verbosity\n"+\
            "  -c, --conf, --config=CONFIG_FILE\tLoad this file on startup instead of .pronsolerc; you may chain config files, if so settings auto-save will use the last specified file\n"+\
            "  -e, --execute=COMMAND\t\tExecutes command after configuration/.pronsolerc is loaded; macros/settings from these commands are not autosaved"

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hVvc:e:", ["help", "version", "verbose", "conf=", "config=", "execute="])
    except getopt.GetoptError as err:
        print(str(err))
        print(usage)
        sys.exit(2)
    for o, a in opts:
        if o in ('-V','--version'):
            print("printrun "+printcore_version)
            sys.exit(0)
        elif o in ('-h', '--help'):
            print(usage)
            sys.exit(0)

    interp = pronsole()
    interp.parse_cmdline(sys.argv[1:])
    try:
        interp.cmdloop()
    except SystemExit:
        interp.p.disconnect()
    except:
        logging.error(_("Caught an exception, exiting:")
                      + "\n" + traceback.format_exc())
        interp.p.disconnect()



# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.

import platform
import logging
import os

if platform.system() == "Darwin":
    from .osx import inhibit_sleep_osx, deinhibit_sleep_osx
    inhibit_sleep = inhibit_sleep_osx
    deinhibit_sleep = deinhibit_sleep_osx
elif platform.system() == "Windows":
    import ctypes
    ES_CONTINUOUS = 0x80000000
    ES_SYSTEM_REQUIRED = 0x00000001

    def inhibit_sleep(reason):
        mode = ES_CONTINUOUS | ES_SYSTEM_REQUIRED
        ctypes.windll.kernel32.SetThreadExecutionState(ctypes.c_int(mode))

    def deinhibit_sleep():
        ctypes.windll.kernel32.SetThreadExecutionState(ctypes.c_int(ES_CONTINUOUS))
else:
    try:
        import dbus
        inhibit_sleep_handler = None
        inhibit_sleep_token = None
        bus = dbus.SessionBus()
        try:
            if os.environ.get('DESKTOP_SESSION') == "mate":
                # Mate uses a special service
                service_name = "org.mate.ScreenSaver"
                object_path = "/org/mate/ScreenSaver"
            else:
                # standard service name
                service_name = "org.freedesktop.ScreenSaver"
                object_path = "/org/freedesktop/ScreenSaver"
            # GNOME and Mate use the right object path, try it first
            proxy = bus.get_object(service_name, object_path)
            inhibit_sleep_handler = dbus.Interface(proxy, service_name)
            # Do a test run
            token = inhibit_sleep_handler.Inhibit("printrun", "test")
            inhibit_sleep_handler.UnInhibit(token)
        except dbus.DBusException:
            # KDE uses /ScreenSaver object path, let's try it as well
            proxy = bus.get_object(service_name,
                                   "/ScreenSaver")
            inhibit_sleep_handler = dbus.Interface(proxy, service_name)
            token = inhibit_sleep_handler.Inhibit("printrun", "test")
            inhibit_sleep_handler.UnInhibit(token)

        def inhibit_sleep(reason):
            global inhibit_sleep_handler, inhibit_sleep_token
            inhibit_sleep_token = inhibit_sleep_handler.Inhibit("printrun", reason)

        def deinhibit_sleep():
            global inhibit_sleep_handler, inhibit_sleep_token
            if inhibit_sleep_handler is None or inhibit_sleep_token is None:
                return
            inhibit_sleep_handler.UnInhibit(inhibit_sleep_token)
            inhibit_sleep_token = None
    except Exception as e:
        logging.warning("Could not setup DBus for sleep inhibition: %s" % e)

        def inhibit_sleep(reason):
            return

        def deinhibit_sleep():
            return

try:
    import psutil

    def get_nice(nice, p = None):
        if not p: p = psutil.Process(os.getpid())
        if callable(p.nice):
            return p.nice()
        else:
            return p.nice

    def set_nice(nice, p = None):
        if not p: p = psutil.Process(os.getpid())
        if callable(p.nice):
            p.nice(nice)
        else:
            p.nice = nice

    if platform.system() != "Windows":
        import resource
        if hasattr(psutil, "RLIMIT_NICE"):
            nice_limit, _ = resource.getrlimit(psutil.RLIMIT_NICE)
            high_priority_nice = 20 - nice_limit
        else:
            high_priority_nice = 0
            # RLIMIT_NICE is not available (probably OSX), let's probe
            # Try setting niceness to -20 .. -1
            p = psutil.Process(os.getpid())
            orig_nice = get_nice(p)
            for i in range(-20, 0):
                try:
                    set_nice(i, p)
                    high_priority_nice = i
                    break
                except psutil.AccessDenied as e:
                    pass
            set_nice(orig_nice, p)

    def set_priority():
        if platform.system() == "Windows":
            set_nice(psutil.HIGH_PRIORITY_CLASS)
        else:
            if high_priority_nice < 0:
                set_nice(high_priority_nice)

    def reset_priority():
        if platform.system() == "Windows":
            set_nice(psutil.NORMAL_PRIORITY_CLASS)
        else:
            if high_priority_nice < 0:
                set_nice(0)

    def powerset_print_start(reason):
        set_priority()
        inhibit_sleep(reason)

    def powerset_print_stop():
        reset_priority()
        deinhibit_sleep()
except ImportError as e:
    logging.warning("psutil unavailable, could not import power utils:" + str(e))

    def powerset_print_start(reason):
        pass

    def powerset_print_stop():
        pass



# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.

# Standard libraries:
import os
import platform
import re
import selectors
import socket
import time

# Third-party libraries
import serial

READ_EMPTY = b''
"""Constant to represent empty or no data"""

READ_EOF = None
"""Constant to represent an end-of-file"""


class Device():
    """Handler for serial and web socket connections.

    Provides the same functions for both so it abstracts what kind of
    connection is being used.

    Parameters
    ----------
    port : str, optional
        Either a device name, such as '/dev/ttyUSB0' or 'COM3', or an URL with
        port, such as '192.168.0.10:80' or 'http://www.example.com:8080'.
    baudrate : int, optional
        Communication speed in bit/s, such as 9600, 115200 or 250000.
        (Default is 9600)
    force_dtr : bool or None, optional
        On serial connections, force the DTR bit to a specific logic level
        (1 or 0) after a successful connection. Not all OS/drivers support
        this functionality. By default it is set to "None" to let the system
        handle it automatically.
    parity_workaround : bool, optional
        On serial connections, enable/disable a workaround on parity
        checking. Not all platforms need to do this parity workaround, and
        some drivers don't support it. By default it is disabled.

    Attributes
    ----------
    is_connected
    has_flow_control

    """

    def __init__(self, port=None, baudrate=9600, force_dtr=None,
                 parity_workaround=False):
        self.port = port
        self.baudrate = baudrate
        self.force_dtr = force_dtr
        self.parity_workaround = parity_workaround

        # Private
        self._device = None
        self._is_connected = False
        self._hostname = None
        self._socketfile = None
        self._port_number = None
        self._read_buffer = []
        self._selector = None
        self._timeout = 0.25
        self._type = None

        if port is not None:
            self._parse_type()

    def connect(self, port=None, baudrate=None):
        """Establishes the connection to the device.

        Parameters
        ----------
        port : str, optional
            See `port` attribute. Only required if it was not provided
            already.
        baudrate : int, optional
            See `baudrate` attribute. Only required if it was not provided
            already.

        Raises
        ------
        DeviceError
            If an error occurred when attempting to connect.

        """
        if port is not None:
            self.port = port
        if baudrate is not None:
            self.baudrate = baudrate

        if self.port is not None:
            self._parse_type()
            getattr(self, "_connect_" + self._type)()
        else:
            raise DeviceError("No port or URL specified")

    def disconnect(self):
        """Terminates the connection to the device."""
        if self._device is not None:
            getattr(self, "_disconnect_" + self._type)()

    @property
    def is_connected(self):
        """True if connection to peer is alive.

        Warnings
        --------
        Current implementation for socket connections only tracks status of
        the connection but does not actually check it. So, if it is used to
        check the connection before sending data, it might fail to prevent an
        error being raised due to a lost connection.

        """
        if self._device is not None:
            return getattr(self, "_is_connected_" + self._type)()
        return False

    @property
    def has_flow_control(self):
        """True if the device has flow control mechanics."""
        if self._type == 'socket':
            return True
        return False

    def readline(self) -> bytes:
        """Read one line from the device stream.

        Returns
        -------
        bytes
            Array containing the feedback received from the
            device. `READ_EMPTY` will be returned if no data was
            available. `READ_EOF` is returned if connection was terminated at
            the other end.

        Raises
        ------
        DeviceError
            If connected peer is unreachable.

        """
        # TODO: silent fail on no device? return timeout?
        if self._device is not None:
            return getattr(self, "_readline_" + self._type)()
        raise DeviceError("Attempted to read when disconnected")

    def reset(self):
        """Attempt to reset the connection to the device.

        Warnings
        --------
        Current implementation has no effect on socket connections.

        """
        if self._device is not None:
            if self._type == 'serial':
                getattr(self, "_reset_" + self._type)()

    def write(self, data: bytes):
        """Write data to the connected peer.

        Parameters
        ----------
        data: bytes
            The bytes data to be written. This should be of type `bytes` (or
            compatible such as `bytearray` or `memoryview`). Unicode strings
            must be encoded.

        Raises
        ------
        DeviceError
            If connected peer is unreachable.
        TypeError
            If `data` is not of 'bytes' type.

        """
        if self._device is not None:
            getattr(self, "_write_" + self._type)(data)
        else:
            raise DeviceError("Attempted to write when disconnected")

    def _parse_type(self):
        # Guess which type of connection is being used
        if self._is_url(self.port):
            self._type = 'socket'
        else:
            self._type = 'serial'

    def _is_url(self, text):
        # TODO: Rearrange to avoid long line
        host_regexp = re.compile("^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$|^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$")
        if ':' in text:
            bits = text.split(":")
            if len(bits) == 2:
                self._hostname = bits[0]
                try:
                    self._port_number = int(bits[1])
                    if (host_regexp.match(self._hostname) and
                            1 <= self._port_number <= 65535):
                        return True
                except:
                    # TODO: avoid catch-all clauses
                    pass
        return False

    # ------------------------------------------------------------------------
    # Serial Functions
    # ------------------------------------------------------------------------
    def _connect_serial(self):
        # Disable HUPCL
        # TODO: Check if still required
        self._disable_ttyhup()

        try:
            # TODO: Check if this trick is still needed
            if self.parity_workaround:
                self._device = serial.Serial(port=self.port,
                                             baudrate=self.baudrate,
                                             timeout=0.25,
                                             parity=serial.PARITY_ODD)
                self._device.close()
                self._device.parity = serial.PARITY_NONE
            else:
                self._device = serial.Serial(baudrate=self.baudrate,
                                             timeout=0.25,
                                             parity=serial.PARITY_NONE)
                self._device.port = self.port

            # TODO: Check if this is still required
            if self.force_dtr is not None:
                self._device.dtr = self.force_dtr

            self._device.open()

        except (serial.SerialException, IOError) as e:
            msg = "Could not connect to serial port '{}'".format(self.port)
            raise DeviceError(msg, e) from e

    def _is_connected_serial(self):
        return self._device.is_open

    def _disconnect_serial(self):
        try:
            self._device.close()
        except serial.SerialException as e:
            msg = "Error on serial disconnection"
            raise DeviceError(msg, e) from e

    def _readline_serial(self):
        try:
            # Serial.readline() returns b'' (aka `READ_EMPTY`) on timeout
            return self._device.readline()
        except (serial.SerialException, OSError) as e:
            msg = f"Unable to read from serial port '{self.port}'"
            raise DeviceError(msg, e) from e

    def _reset_serial(self):
        self._device.dtr = True
        time.sleep(0.2)
        self._device.dtr = False

    def _write_serial(self, data):
        try:
            self._device.write(data)
        except serial.SerialException as e:
            msg = "Unable to write to serial port '{self.port}'"
            raise DeviceError(msg, e) from e

    def _disable_ttyhup(self):
        if platform.system() == "Linux":
            os.system("stty -F %s -hup" % self.port)

    # ------------------------------------------------------------------------
    # Socket Functions
    # ------------------------------------------------------------------------
    def _connect_socket(self):
        self._device = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._device.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self._timeout = 0.25
        self._device.settimeout(1.0)

        try:
            self._device.connect((self._hostname, self._port_number))
            # A single read timeout raises OSError for all later reads
            # probably since python 3.5 use non blocking instead
            self._device.settimeout(0)
            self._socketfile = self._device.makefile('rwb', buffering=0)
            self._selector = selectors.DefaultSelector()
            self._selector.register(self._device, selectors.EVENT_READ)
            self._is_connected = True

        except OSError as e:
            self._disconnect_socket()
            msg = "Could not connect to {}:{}".format(self._hostname,
                                                      self._port_number)
            raise DeviceError(msg, e) from e

    def _is_connected_socket(self):
        # TODO: current implementation tracks status of connection but
        # does not actually check it. Ref. is_connected()
        return self._is_connected

    def _disconnect_socket(self):
        self._is_connected = False
        try:
            if self._socketfile is not None:
                self._socketfile.close()
            if self._selector is not None:
                self._selector.unregister(self._device)
                self._selector.close()
                self._selector = None
            self._device.close()
        except OSError as e:
            msg = "Error on socket disconnection"
            raise DeviceError(msg, e) from e

    def _readline_socket(self):
        SYS_AGAIN = None  # python's marker for timeout/no data
        # SYS_EOF = b''   # python's marker for EOF
        try:
            line = self._readline_buf()
            if line:
                return line
            chunk_size = 256
            while True:
                chunk = self._socketfile.read(chunk_size)
                if (chunk is SYS_AGAIN and
                        self._selector.select(self._timeout)):
                    chunk = self._socketfile.read(chunk_size)
                if chunk:
                    self._read_buffer.append(chunk)
                    line = self._readline_buf()
                    if line:
                        return line
                elif chunk is SYS_AGAIN:
                    return READ_EMPTY
                else:  # chunk is SYS_EOF
                    line = b''.join(self._read_buffer)
                    self._read_buffer = []
                    if line:
                        return line
                    self._is_connected = False
                    return READ_EOF
        except OSError as e:
            self._is_connected = False
            msg = ("Unable to read from {}:{}. Connection lost"
                   ).format(self._hostname, self._port_number)
            raise DeviceError(msg, e) from e

    def _readline_buf(self):
        # Try to readline from buffer
        if self._read_buffer:
            chunk = self._read_buffer[-1]
            eol = chunk.find(b'\n')
            if eol >= 0:
                line = b''.join(self._read_buffer[:-1]) + chunk[:(eol+1)]
                self._read_buffer = []
                if eol + 1 < len(chunk):
                    self._read_buffer.append(chunk[(eol+1):])
                return line
        return READ_EMPTY

    def _write_socket(self, data):
        try:
            self._socketfile.write(data)
            try:
                self._socketfile.flush()
            except socket.timeout:
                pass
        except (OSError, RuntimeError) as e:
            self._is_connected = False
            msg = ("Unable to write to {}:{}. Connection lost"
                   ).format(self._hostname, self._port_number)
            raise DeviceError(msg, e) from e


class DeviceError(Exception):
    """Raised on any connection error.

    One exception groups all connection errors regardless of the underlying
    connection or error type.

    Parameters
    ----------
    msg : str
        Error message.
    cause : Exception, optional
        Underlying error.

    Attributes
    ----------
    cause

    """

    def __init__(self, msg, cause=None):
        super().__init__(msg)
        self.cause = cause



# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.

import wx

from .utils import make_button

class LogPane(wx.BoxSizer):

    def __init__(self, root, parentpanel = None):
        super(LogPane, self).__init__(wx.VERTICAL)
        if not parentpanel: parentpanel = root.panel
        root.logbox = wx.TextCtrl(parentpanel, style = wx.TE_MULTILINE, size = (350, -1))
        root.logbox.SetMinSize((100, -1))
        root.logbox.SetEditable(0)
        self.Add(root.logbox, 1, wx.EXPAND)
        bottom_panel = root.newPanel(parentpanel)
        lbrs = wx.BoxSizer(wx.HORIZONTAL)
        root.commandbox = wx.TextCtrl(bottom_panel, style = wx.TE_PROCESS_ENTER)
        root.commandbox.SetToolTip(wx.ToolTip(_("Send commands to printer\n(Type 'help' for simple\nhelp function)")))
        root.commandbox.Hint = _("Command to send")
        root.commandbox.Bind(wx.EVT_TEXT_ENTER, root.sendline)
        root.commandbox.Bind(wx.EVT_CHAR, root.cbkey)
        def deselect(ev):
            # In Ubuntu 19.10, when focused, all text is selected
            lp = root.commandbox.LastPosition
            # print(f"SetSelection({lp}, {lp})")
            wx.CallAfter(root.commandbox.SetSelection, lp, lp)
            ev.Skip()
        root.commandbox.Bind(wx.EVT_SET_FOCUS, deselect)
        root.commandbox.history = [""]
        root.commandbox.histindex = 1
        lbrs.Add(root.commandbox, 1)
        root.sendbtn = make_button(bottom_panel, _("Send"), root.sendline, _("Send Command to Printer"), style = wx.BU_EXACTFIT, container = lbrs)
        bottom_panel.SetSizer(lbrs)
        self.Add(bottom_panel, 0, wx.EXPAND)



# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.

import wx

from .utils import make_autosize_button

def MainToolbar(root, parentpanel = None, use_wrapsizer = False):
    if not parentpanel: parentpanel = root.panel
    if root.settings.lockbox:
        root.locker = wx.CheckBox(parentpanel, label = _("Lock") + "  ")
        root.locker.Bind(wx.EVT_CHECKBOX, root.lock)
        root.locker.SetToolTip(wx.ToolTip(_("Lock graphical interface")))
        glob = wx.BoxSizer(wx.HORIZONTAL)
        parentpanel = root.newPanel(parentpanel)
        glob.Add(parentpanel, 1, flag = wx.EXPAND)
        glob.Add(root.locker, 0, flag = wx.ALIGN_CENTER)
    ToolbarSizer = wx.WrapSizer if use_wrapsizer else wx.BoxSizer
    self = ToolbarSizer(wx.HORIZONTAL)
    root.rescanbtn = make_autosize_button(parentpanel, _("Port"), root.rescanports, _("Communication Settings\nClick to rescan ports"))
    self.Add(root.rescanbtn, 0, wx.TOP | wx.LEFT, 0)

    root.serialport = wx.ComboBox(parentpanel, -1, choices = root.scanserial(),
                                  style = wx.CB_DROPDOWN)
    root.serialport.SetToolTip(wx.ToolTip(_("Select Port Printer is connected to")))
    root.rescanports()
    self.Add(root.serialport)

    self.Add(wx.StaticText(parentpanel, -1, "@"), 0, wx.RIGHT | wx.ALIGN_CENTER, 0)
    root.baud = wx.ComboBox(parentpanel, -1,
                            choices = ["2400", "9600", "19200", "38400",
                                       "57600", "115200", "250000", "500000", "1000000"],
                            style = wx.CB_DROPDOWN, size = (110, -1))
    root.baud.SetToolTip(wx.ToolTip(_("Select Baud rate for printer communication")))
    try:
        root.baud.SetValue("115200")
        root.baud.SetValue(str(root.settings.baudrate))
    except:
        pass
    self.Add(root.baud)

    if not hasattr(root, "connectbtn"):
        root.connectbtn_cb_var = root.connect
        root.connectbtn = make_autosize_button(parentpanel, _("&Connect"), root.connectbtn_cb, _("Connect to the printer"))
        root.statefulControls.append(root.connectbtn)
    else:
        root.connectbtn.Reparent(parentpanel)
    self.Add(root.connectbtn)
    if not hasattr(root, "resetbtn"):
        root.resetbtn = make_autosize_button(parentpanel, _("Reset"), root.reset, _("Reset the printer"))
        root.statefulControls.append(root.resetbtn)
    else:
        root.resetbtn.Reparent(parentpanel)
    self.Add(root.resetbtn)

    self.AddStretchSpacer(prop = 1)

    root.loadbtn = make_autosize_button(parentpanel, _("Load file"), root.loadfile, _("Load a 3D model file"), self)
    root.sdbtn = make_autosize_button(parentpanel, _("SD"), root.sdmenu, _("SD Card Printing"), self)
    root.sdbtn.Reparent(parentpanel)
    root.printerControls.append(root.sdbtn)
    if not hasattr(root, "printbtn"):
        root.printbtn = make_autosize_button(parentpanel, _("Print"), root.printfile, _("Start Printing Loaded File"))
        root.statefulControls.append(root.printbtn)
    else:
        root.printbtn.Reparent(parentpanel)
    self.Add(root.printbtn)
    if not hasattr(root, "pausebtn"):
        root.pausebtn = make_autosize_button(parentpanel, _("Pause"), root.pause, _("Pause Current Print"))
        root.statefulControls.append(root.pausebtn)
    else:
        root.pausebtn.Reparent(parentpanel)
    self.Add(root.pausebtn)
    root.offbtn = make_autosize_button(parentpanel, _("Off"), root.off, _("Turn printer off"), self)
    root.printerControls.append(root.offbtn)

    self.AddStretchSpacer(prop = 4)

    if root.settings.lockbox:
        parentpanel.SetSizer(self)
        return glob
    else:
        return self



#!/usr/bin/env python3

# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.

import logging
import wx

from . import gcoder
from .gl.panel import wxGLPanel
from .gl.trackball import build_rotmatrix
from .gl.libtatlin import actors
from .injectgcode import injector, injector_edit

from pyglet.gl import glPushMatrix, glPopMatrix, \
    glTranslatef, glRotatef, glScalef, glMultMatrixd, \
    glGetDoublev, GL_MODELVIEW_MATRIX, GLdouble

from .gviz import GvizBaseFrame

from .gui.widgets import get_space
from .utils import imagefile, install_locale, get_home_pos
install_locale('pronterface')

def create_model(light):
    if light:
        return actors.GcodeModelLight()
    else:
        return actors.GcodeModel()

def gcode_dims(g):
    return ((g.xmin, g.xmax, g.width),
            (g.ymin, g.ymax, g.depth),
            (g.zmin, g.zmax, g.height))

def set_model_colors(model, root):
    for field in dir(model):
        if field.startswith("color_"):
            root_fieldname = "gcview_" + field
            if hasattr(root, root_fieldname):
                setattr(model, field, getattr(root, root_fieldname))

def recreate_platform(self, build_dimensions, circular, grid):
    self.platform = actors.Platform(build_dimensions, circular = circular, grid = grid)
    self.objects[0].model = self.platform
    wx.CallAfter(self.Refresh)

def set_gcview_params(self, path_width, path_height):
    self.path_halfwidth = path_width / 2
    self.path_halfheight = path_height / 2
    has_changed = False
    for obj in self.objects[1:]:
        if isinstance(obj.model, actors.GcodeModel):
            obj.model.set_path_size(self.path_halfwidth, self.path_halfheight)
            has_changed = True
    return has_changed

# E selected for Up because is above D
LAYER_UP_KEYS = ord('U'), ord('E'), wx.WXK_UP
LAYER_DOWN_KEYS = ord('D'), wx.WXK_DOWN
ZOOM_IN_KEYS = wx.WXK_PAGEDOWN, 388, wx.WXK_RIGHT, ord('+')
ZOOM_OUT_KEYS = wx.WXK_PAGEUP, 390, wx.WXK_LEFT, ord('-')
FIT_KEYS = [ord('F')]
CURRENT_LAYER_KEYS = [ord('C')]
RESET_KEYS = [ord('R')]

class GcodeViewPanel(wxGLPanel):

    def __init__(self, parent,
                 build_dimensions = (200, 200, 100, 0, 0, 0),
                 realparent = None, antialias_samples = 0, perspective=False):
        if perspective:
            self.orthographic=False
        super().__init__(parent, wx.DefaultPosition,
                                             wx.DefaultSize, 0,
                                             antialias_samples = antialias_samples)
        self.canvas.Bind(wx.EVT_MOUSE_EVENTS, self.move)
        self.canvas.Bind(wx.EVT_LEFT_DCLICK, self.double)
        # self.canvas.Bind(wx.EVT_KEY_DOWN, self.keypress)
        # in Windows event inspector shows only EVT_CHAR_HOOK events
        self.canvas.Bind(wx.EVT_CHAR_HOOK, self.keypress)
        self.initialized = 0
        self.canvas.Bind(wx.EVT_MOUSEWHEEL, self.wheel)
        self.parent = realparent or parent
        self.initpos = None
        self.build_dimensions = build_dimensions
        self.dist = max(self.build_dimensions[:2])
        self.basequat = [0, 0, 0, 1]
        self.mousepos = [0, 0]

    def inject(self):
        l = self.parent.model.num_layers_to_draw
        filtered = [k for k, v in self.parent.model.layer_idxs_map.items() if v == l]
        if filtered:
            injector(self.parent.model.gcode, l, filtered[0])
        else:
            logging.error(_("Invalid layer for injection"))

    def editlayer(self):
        l = self.parent.model.num_layers_to_draw
        filtered = [k for k, v in self.parent.model.layer_idxs_map.items() if v == l]
        if filtered:
            injector_edit(self.parent.model.gcode, l, filtered[0])
        else:
            logging.error(_("Invalid layer for edition"))

    def setlayercb(self, layer):
        pass

    def OnInitGL(self, *args, **kwargs):
        super().OnInitGL(*args, **kwargs)
        filenames = getattr(self.parent, 'filenames', None)
        if filenames:
            for filename in filenames:
                self.parent.load_file(filename)
            self.parent.autoplate()
            getattr(self.parent, 'loadcb', bool)()
            self.parent.filenames = None

    def create_objects(self):
        '''create opengl objects when opengl is initialized'''
        for obj in self.parent.objects:
            if obj.model and obj.model.loaded and not obj.model.initialized:
                obj.model.init()

    def update_object_resize(self):
        '''called when the window receives only if opengl is initialized'''
        pass

    def draw_objects(self):
        '''called in the middle of ondraw after the buffer has been cleared'''
        self.create_objects()

        glPushMatrix()
        # Rotate according to trackball
        glMultMatrixd(build_rotmatrix(self.basequat))
        # Move origin to bottom left of platform
        platformx0 = -self.build_dimensions[3] - self.parent.platform.width / 2
        platformy0 = -self.build_dimensions[4] - self.parent.platform.depth / 2
        glTranslatef(platformx0, platformy0, 0)

        for obj in self.parent.objects:
            if not obj.model \
               or not obj.model.loaded:
                continue
            # Skip (comment out) initialized check, which safely causes empty
            # model during progressive load. This can cause exceptions/garbage
            # render, but seems fine for now
            # May need to lock init() and draw_objects() together
            # if not obj.model.initialized:
            #     continue
            glPushMatrix()
            glTranslatef(*(obj.offsets))
            glRotatef(obj.rot, 0.0, 0.0, 1.0)
            glTranslatef(*(obj.centeroffset))
            glScalef(*obj.scale)

            obj.model.display()
            glPopMatrix()
        glPopMatrix()

    # ==========================================================================
    # Utils
    # ==========================================================================
    def get_modelview_mat(self, local_transform):
        mvmat = (GLdouble * 16)()
        if local_transform:
            glPushMatrix()
            # Rotate according to trackball
            glMultMatrixd(build_rotmatrix(self.basequat))
            # Move origin to bottom left of platform
            platformx0 = -self.build_dimensions[3] - self.parent.platform.width / 2
            platformy0 = -self.build_dimensions[4] - self.parent.platform.depth / 2
            glTranslatef(platformx0, platformy0, 0)
            glGetDoublev(GL_MODELVIEW_MATRIX, mvmat)
            glPopMatrix()
        else:
            glGetDoublev(GL_MODELVIEW_MATRIX, mvmat)
        return mvmat

    def double(self, event):
        getattr(self.parent, 'clickcb', bool)(event)

    def move(self, event):
        """react to mouse actions:
        no mouse: show red mousedrop
        LMB: rotate viewport
        RMB: move viewport
        """
        if event.Entering():
            self.canvas.SetFocus()
            event.Skip()
            return
        if event.Dragging():
            if event.LeftIsDown():
                self.handle_rotation(event)
            elif event.RightIsDown():
                self.handle_translation(event)
            self.Refresh(False)
        elif event.LeftUp() or event.RightUp():
            self.initpos = None
        event.Skip()

    def layerup(self):
        if not getattr(self.parent, 'model', False):
            return
        max_layers = self.parent.model.max_layers
        current_layer = self.parent.model.num_layers_to_draw
        # accept going up to max_layers + 1
        # max_layers means visualizing the last layer differently,
        # max_layers + 1 means visualizing all layers with the same color
        new_layer = min(max_layers + 1, current_layer + 1)
        self.parent.model.num_layers_to_draw = new_layer
        self.parent.setlayercb(new_layer)
        wx.CallAfter(self.Refresh)

    def layerdown(self):
        if not getattr(self.parent, 'model', False):
            return
        current_layer = self.parent.model.num_layers_to_draw
        new_layer = max(1, current_layer - 1)
        self.parent.model.num_layers_to_draw = new_layer
        self.parent.setlayercb(new_layer)
        wx.CallAfter(self.Refresh)

    wheelTimestamp = None
    def handle_wheel(self, event):
        if self.wheelTimestamp == event.Timestamp:
            # filter duplicate event delivery in Ubuntu, Debian issue #1110
            return

        self.wheelTimestamp = event.Timestamp

        delta = event.GetWheelRotation()
        factor = 1.05
        if event.ControlDown():
            factor = 1.02
        if hasattr(self.parent, "model") and event.ShiftDown():
            if not self.parent.model:
                return
            count = 1 if not event.ControlDown() else 10
            for i in range(count):
                if delta > 0:
                    self.layerup()
                else:
                    self.layerdown()
            return
        x, y = event.GetPosition() * self.GetContentScaleFactor()
        x, y, _ = self.mouse_to_3d(x, y)
        if delta > 0:
            self.zoom(factor, (x, y))
        else:
            self.zoom(1 / factor, (x, y))

    def wheel(self, event):
        """react to mouse wheel actions:
            without shift: set max layer
            with shift: zoom viewport
        """
        self.handle_wheel(event)
        wx.CallAfter(self.Refresh)

    def fit(self):
        if not self.parent.model or not self.parent.model.loaded:
            return
        self.canvas.SetCurrent(self.context)
        dims = gcode_dims(self.parent.model.gcode)
        self.reset_mview(1.0)
        center_x = (dims[0][0] + dims[0][1]) / 2
        center_y = (dims[1][0] + dims[1][1]) / 2
        center_x = self.build_dimensions[0] / 2 - center_x
        center_y = self.build_dimensions[1] / 2 - center_y
        if self.orthographic:
            ratio = float(self.dist) / max(dims[0][2], dims[1][2])
            glScalef(ratio, ratio, 1)
        glTranslatef(center_x, center_y, 0)
        wx.CallAfter(self.Refresh)

    def keypress(self, event):
        """gets keypress events and moves/rotates active shape"""
        if event.HasModifiers():
            # let alt+c bubble up
            event.Skip()
            return
        step = event.ControlDown() and 1.05 or 1.1
        key = event.GetKeyCode()
        if key in LAYER_UP_KEYS:
            self.layerup()
            return  # prevent shifting focus to other controls
        elif key in LAYER_DOWN_KEYS:
            self.layerdown()
            return
        # x, y, _ = self.mouse_to_3d(self.width / 2, self.height / 2)
        elif key in ZOOM_IN_KEYS:
            self.zoom_to_center(step)
            return
        elif key in ZOOM_OUT_KEYS:
            self.zoom_to_center(1 / step)
            return
        elif key in FIT_KEYS:
            self.fit()
        elif key in CURRENT_LAYER_KEYS:
            if not self.parent.model or not self.parent.model.loaded:
                return
            self.parent.model.only_current = not self.parent.model.only_current
            wx.CallAfter(self.Refresh)
        elif key in RESET_KEYS:
            self.resetview()
        event.Skip()

    def resetview(self):
        self.canvas.SetCurrent(self.context)
        self.reset_mview(0.9)
        self.basequat = [0, 0, 0, 1]
        wx.CallAfter(self.Refresh)

class GCObject:

    def __init__(self, model):
        self.offsets = [0, 0, 0]
        self.centeroffset = [0, 0, 0]
        self.rot = 0
        self.curlayer = 0.0
        self.scale = [1.0, 1.0, 1.0]
        self.model = model

class GcodeViewLoader:

    path_halfwidth = 0.2
    path_halfheight = 0.15

    def addfile_perlayer(self, gcode = None, showall = False):
        self.model = create_model(self.root.settings.light3d
                                  if self.root else False)
        if isinstance(self.model, actors.GcodeModel):
            self.model.set_path_size(self.path_halfwidth, self.path_halfheight)
        self.objects[-1].model = self.model
        if self.root:
            set_model_colors(self.model, self.root)
        if gcode is not None:
            generator = self.model.load_data(gcode)
            generator_output = next(generator)
            while generator_output is not None:
                yield generator_output
                generator_output = next(generator)
        wx.CallAfter(self.Refresh)
        yield None

    def addfile(self, gcode = None, showall = False):
        generator = self.addfile_perlayer(gcode, showall)
        while next(generator) is not None:
            continue

    def set_gcview_params(self, path_width, path_height):
        return set_gcview_params(self, path_width, path_height)

from printrun.gviz import BaseViz
class GcodeViewMainWrapper(GcodeViewLoader, BaseViz):

    def __init__(self, parent, build_dimensions, root, circular, antialias_samples, grid, perspective=False):
        self.root = root
        self.glpanel = GcodeViewPanel(parent, realparent = self,
                                      build_dimensions = build_dimensions,
                                      antialias_samples = antialias_samples, perspective=perspective)
        self.glpanel.SetMinSize((150, 150))
        if self.root and hasattr(self.root, "gcview_color_background"):
            self.glpanel.color_background = self.root.gcview_color_background
        self.clickcb = None
        self.widget = self.glpanel
        self.refresh_timer = wx.CallLater(100, self.Refresh)
        self.p = self  # Hack for backwards compatibility with gviz API
        self.grid = grid
        self.platform = actors.Platform(build_dimensions, circular = circular, grid = grid)
        self.model = None
        self.objects = [GCObject(self.platform), GCObject(None)]

    def __getattr__(self, name):
        return getattr(self.glpanel, name)

    def on_settings_change(self, changed_settings):
        if self.model:
            for s in changed_settings:
                if s.name.startswith('gcview_color_'):
                    self.model.update_colors()
                    break

    def set_current_gline(self, gline):
        if gline.is_move and gline.gcview_end_vertex is not None \
           and self.model and self.model.loaded:
            self.model.printed_until = gline.gcview_end_vertex
            if not self.refresh_timer.IsRunning():
                self.refresh_timer.Start()

    def recreate_platform(self, build_dimensions, circular, grid):
        return recreate_platform(self, build_dimensions, circular, grid)

    def setlayer(self, layer):
        if layer in self.model.layer_idxs_map:
            viz_layer = self.model.layer_idxs_map[layer]
            self.parent.model.num_layers_to_draw = viz_layer
            wx.CallAfter(self.Refresh)

    def clear(self):
        self.model = None
        self.objects[-1].model = None
        wx.CallAfter(self.Refresh)

class GcodeViewFrame(GvizBaseFrame, GcodeViewLoader):
    '''A simple class for using OpenGL with wxPython.'''

    def __init__(self, parent, ID, title, build_dimensions, objects = None,
                 pos = wx.DefaultPosition, size = wx.DefaultSize,
                 style = wx.DEFAULT_FRAME_STYLE, root = None, circular = False,
                 antialias_samples = 0,
                 grid = (1, 10), perspective=False):
        GvizBaseFrame.__init__(self, parent, ID, title,
                               pos, size, style)
        self.root = root

        panel, h_sizer = self.create_base_ui()

        self.refresh_timer = wx.CallLater(100, self.Refresh)
        self.p = self  # Hack for backwards compatibility with gviz API
        self.clonefrom = objects
        self.platform = actors.Platform(build_dimensions, circular = circular, grid = grid)
        self.model = objects[1].model if objects else None
        self.objects = [GCObject(self.platform), GCObject(None)]

        fit_image = wx.Image(imagefile('fit.png'), wx.BITMAP_TYPE_PNG).ConvertToBitmap()
        tool_pos = self.toolbar.GetToolPos(3) + 1
        self.toolbar.InsertTool(tool_pos, 10, " " + _("Fit to view"), fit_image,
                                shortHelp = _("Fit to view [F]"), longHelp = _("Fit view to display entire print"))
        self.toolbar.Realize()
        self.glpanel = GcodeViewPanel(panel,
                                      build_dimensions = build_dimensions,
                                      realparent = self,
                                      antialias_samples = antialias_samples, perspective=perspective)

        if self.root and hasattr(self.root, "gcview_color_background"):
            self.glpanel.color_background = self.root.gcview_color_background

        h_sizer.Add(self.glpanel, 1, wx.EXPAND)
        h_sizer.Add(self.layerslider, 0, wx.EXPAND | wx.ALL, get_space('minor'))
        self.glpanel.SetToolTip(_("Click [Mouse-Right] to move the view, click [Mouse-Left] to pan the view, scroll with [Mouse Wheel] to zoom, and scroll with [Shift]+[Mouse Wheel] to change the layer."))

        minsize = self.toolbar.GetEffectiveMinSize().width
        self.SetMinClientSize((minsize, minsize))

        self.Bind(wx.EVT_TOOL, lambda x: self.glpanel.zoom_to_center(1 / 1.2), id = 1)
        self.Bind(wx.EVT_TOOL, lambda x: self.glpanel.zoom_to_center(1.2), id = 2)
        self.Bind(wx.EVT_TOOL, lambda x: self.glpanel.resetview(), id = 3)
        self.Bind(wx.EVT_TOOL, lambda x: self.glpanel.layerdown(), id = 4)
        self.Bind(wx.EVT_TOOL, lambda x: self.glpanel.layerup(), id = 5)
        self.Bind(wx.EVT_TOOL, lambda x: self.glpanel.fit(), id = 10)
        self.Bind(wx.EVT_TOOL, lambda x: self.glpanel.inject(), id = 6)
        self.Bind(wx.EVT_TOOL, lambda x: self.glpanel.editlayer(), id = 7)
        self.Bind(wx.EVT_TOOL, lambda x: self.Close(), id = 9)

    def setlayercb(self, layer):
        self.layerslider.SetValue(layer)
        self.update_status("")

    def update_status(self, extra):
        layer = self.model.num_layers_to_draw
        filtered = [k for k, v in self.model.layer_idxs_map.items() if v == layer]
        if filtered:
            true_layer = filtered[0]
            z = self.model.gcode.all_layers[true_layer].z
            message = _("Layer %d: %s Z = %.03f mm") % (layer, extra, z)
        else:
            message = _("Last layer: Object complete")
        wx.CallAfter(self.SetStatusText, message, 0)

    def process_slider(self, event):
        if self.model is not None:
            new_layer = self.layerslider.GetValue()
            new_layer = min(self.model.max_layers + 1, new_layer)
            new_layer = max(1, new_layer)
            self.model.num_layers_to_draw = new_layer
            self.update_status("")
            wx.CallAfter(self.Refresh)
        else:
            logging.info(_("G-Code Viewer: Can't process slider. Please wait until model is loaded completely."))

    def set_current_gline(self, gline):
        if gline.is_move and gline.gcview_end_vertex is not None \
           and self.model and self.model.loaded:
            self.model.printed_until = gline.gcview_end_vertex
            if not self.refresh_timer.IsRunning():
                self.refresh_timer.Start()

    def recreate_platform(self, build_dimensions, circular, grid):
        return recreate_platform(self, build_dimensions, circular, grid)

    def addfile(self, gcode = None):
        if self.clonefrom:
            self.model = self.clonefrom[-1].model.copy()
            self.objects[-1].model = self.model
        else:
            GcodeViewLoader.addfile(self, gcode)
        self.layerslider.SetRange(1, self.model.max_layers + 1)
        self.layerslider.SetValue(self.model.max_layers + 1)
        wx.CallAfter(self.SetStatusText, _("Last layer: Object complete"), 0)
        wx.CallAfter(self.Refresh)

    def clear(self):
        self.model = None
        self.objects[-1].model = None
        wx.CallAfter(self.Refresh)

if __name__ == "__main__":
    import sys
    app = wx.App(redirect = False)
    build_dimensions = [200, 200, 100, 0, 0, 0]
    title = _("G-Code Viewer")
    frame = GcodeViewFrame(None, wx.ID_ANY, title, size = (400, 400),
                           build_dimensions = build_dimensions)
    gcode = gcoder.GCode(open(sys.argv[1]), get_home_pos(build_dimensions))
    frame.addfile(gcode)

    first_move = None
    for i in range(len(gcode.lines)):
        if gcode.lines[i].is_move:
            first_move = gcode.lines[i]
            break
    last_move = None
    for i in range(len(gcode.lines) - 1, -1, -1):
        if gcode.lines[i].is_move:
            last_move = gcode.lines[i]
            break
    nsteps = 20
    steptime = 500
    lines = [first_move] + [gcode.lines[int(float(i) * (len(gcode.lines) - 1) / nsteps)] for i in range(1, nsteps)] + [last_move]
    current_line = 0

    def setLine():
        global current_line
        frame.set_current_gline(lines[current_line])
        current_line = (current_line + 1) % len(lines)
        timer.Start()
    timer = wx.CallLater(steptime, setLine)
    timer.Start()

    frame.Show(True)
    app.MainLoop()
    app.Destroy()



# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.

import wx
import math
from .bufferedcanvas import BufferedCanvas
from printrun.utils import imagefile

def sign(n):
    if n < 0: return -1
    elif n > 0: return 1
    else: return 0

DASHES = [4, 7]
# Brush and pen for grey overlay when mouse hovers over
HOVER_PEN_COLOR = wx.Colour(100, 100, 100, 172)
HOVER_BRUSH_COLOR = wx.Colour(0, 0, 0, 128)

class FocusCanvas(BufferedCanvas):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.Bind(wx.EVT_SET_FOCUS, self.onFocus)
    
    def onFocus(self, evt):
        self.Refresh()
        evt.Skip()

    def drawFocusRect(self, dc):
        if self.HasFocus():
            pen = wx.Pen(wx.BLACK, 1, wx.PENSTYLE_USER_DASH)
            pen.SetDashes(DASHES)
            dc.Pen = pen
            dc.Brush = wx.Brush(wx.TRANSPARENT_BRUSH)
            dc.DrawRectangle(self.ClientRect)

class XYButtons(FocusCanvas):
    keypad_positions = {
        0: (104, 99),
        1: (86, 83),
        2: (68, 65),
        3: (53, 50)
    }
    keypad_radius = 9
    corner_size = (49, 49)
    corner_inset = (7, 13)
    label_overlay_positions = {
        1: (145, 98.5, 9),
        2: (160.5, 83.5, 10.6),
        3: (178, 66, 13),
        4: (197.3, 46.3, 13.3)
    }
    concentric_circle_radii = [0, 17, 45, 69, 94, 115]
    concentric_inset = 11
    center = (124, 121)
    spacer = 7
    imagename = "control_xy.png"
    corner_to_axis = {
        -1: "center",
        0: "x",
        1: "z",
        2: "y",
        3: "all",
    }

    def __init__(self, parent, moveCallback = None, cornerCallback = None, spacebarCallback = None, bgcolor = "#FFFFFF", ID=-1, zcallback=None):
        self.bg_bmp = wx.Image(imagefile(self.imagename), wx.BITMAP_TYPE_PNG).ConvertToBitmap()
        self.keypad_bmp = wx.Image(imagefile("arrow_keys.png"), wx.BITMAP_TYPE_PNG).ConvertToBitmap()
        self.keypad_idx = -1
        self.hovered_keypad = None
        self.quadrant = None
        self.concentric = None
        self.corner = None
        self.moveCallback = moveCallback
        self.cornerCallback = cornerCallback
        self.spacebarCallback = spacebarCallback
        self.zCallback = zcallback
        self.enabled = False
        # Remember the last clicked buttons, so we can repeat when spacebar pressed
        self.lastMove = None
        self.lastCorner = None

        self.bgcolor = wx.Colour()
        self.bgcolor.Set(bgcolor)
        self.bgcolormask = wx.Colour(self.bgcolor.Red(), self.bgcolor.Green(), self.bgcolor.Blue(), 128)

        super().__init__(parent, ID, size=self.bg_bmp.GetSize())

        self.bind_events()

    def bind_events(self):
        # Set up mouse and keyboard event capture
        self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnLeftDown)
        self.Bind(wx.EVT_MOTION, self.OnMotion)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.OnLeaveWindow)
        self.Bind(wx.EVT_CHAR_HOOK, self.OnKey)
        self.Bind(wx.EVT_KILL_FOCUS, self.onKillFocus)

    def onKillFocus(self, evt):
        self.setKeypadIndex(-1)
        evt.Skip()

    def disable(self):
        self.Enabled = self.enabled = False
        self.update()

    def enable(self):
        self.Enabled = self.enabled = True
        self.update()

    def repeatLast(self):
        if self.lastMove:
            self.moveCallback(*self.lastMove)
        if self.lastCorner:
            self.cornerCallback(self.corner_to_axis[self.lastCorner])

    def clearRepeat(self):
        self.lastMove = None
        self.lastCorner = None

    def distanceToLine(self, pos, x1, y1, x2, y2):
        xlen = x2 - x1
        ylen = y2 - y1
        pxlen = x1 - pos.x
        pylen = y1 - pos.y
        return abs(xlen * pylen - ylen * pxlen) / math.sqrt(xlen ** 2 + ylen ** 2)

    def distanceToPoint(self, x1, y1, x2, y2):
        return math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

    def cycleKeypadIndex(self, forward):
        idx = self.keypad_idx + (1 if forward else -1)
        # do not really cycle to allow exiting of jog controls widget
        return idx if idx < len(self.keypad_positions) else -1

    def setKeypadIndex(self, idx):
        self.keypad_idx = idx
        self.update()

    def getMovement(self, event):
        xdir = [1, 0, -1, 0, 0, 0][self.quadrant]
        ydir = [0, 1, 0, -1, 0, 0][self.quadrant]
        zdir = [0, 0, 0, 0, 1, -1][self.quadrant]
        magnitude = math.pow(10, self.concentric - 2)
        magnitude *= event.ShiftDown() and 2 or event.ControlDown() and 0.5 or 1

        if zdir:
            magnitude = min(magnitude, 10)
        return (magnitude * xdir, magnitude * ydir, magnitude * zdir)

    def lookupConcentric(self, radius):
        idx = 0
        for r in self.concentric_circle_radii[1:]:
            if radius < r:
                return idx
            idx += 1
        return len(self.concentric_circle_radii)

    def getQuadrantConcentricFromPosition(self, pos):
        rel_x = pos[0] - self.center[0]
        rel_y = pos[1] - self.center[1]
        if rel_x > rel_y and rel_x > -rel_y:
            quadrant = 0  # Right
        elif rel_x <= rel_y and rel_x > -rel_y:
            quadrant = 3  # Down
        elif rel_x > rel_y and rel_x < -rel_y:
            quadrant = 1  # Up
        else:
            quadrant = 2  # Left

        radius = math.sqrt(rel_x ** 2 + rel_y ** 2)
        idx = self.lookupConcentric(radius)
        return (quadrant, idx)

    def mouseOverKeypad(self, mpos):
        for idx, kpos in self.keypad_positions.items():
            radius = self.distanceToPoint(mpos[0], mpos[1], kpos[0], kpos[1])
            if radius < XYButtons.keypad_radius:
                return idx
        return None

    def drawPartialPie(self, gc, center, r1, r2, angle1, angle2):
        p1 = wx.Point(int(center.x + r1 * math.cos(angle1)), int(center.y + r1 * math.sin(angle1)))

        path = gc.CreatePath()
        path.MoveToPoint(p1.x, p1.y)
        path.AddArc(center.x, center.y, r1, angle1, angle2, True)
        path.AddArc(center.x, center.y, r2, angle2, angle1, False)
        path.AddLineToPoint(p1.x, p1.y)
        gc.DrawPath(path)

    def highlightQuadrant(self, gc, quadrant, concentric):
        if not 0 <= quadrant <= 3:
            return
        assert(concentric >= 0 and concentric <= 4)

        inner_ring_radius = self.concentric_inset
        # fudge = math.pi*0.002
        fudge = -0.02
        center = wx.Point(self.center[0], self.center[1])
        if quadrant == 0:
            a1, a2 = (-math.pi * 0.25, math.pi * 0.25)
            center.x += inner_ring_radius
        elif quadrant == 1:
            a1, a2 = (math.pi * 1.25, math.pi * 1.75)
            center.y -= inner_ring_radius
        elif quadrant == 2:
            a1, a2 = (math.pi * 0.75, math.pi * 1.25)
            center.x -= inner_ring_radius
        elif quadrant == 3:
            a1, a2 = (math.pi * 0.25, math.pi * 0.75)
            center.y += inner_ring_radius

        r1 = self.concentric_circle_radii[concentric]
        r2 = self.concentric_circle_radii[concentric + 1]

        self.drawPartialPie(gc, center, r1 - inner_ring_radius, r2 - inner_ring_radius, a1 + fudge, a2 - fudge)

    def drawCorner(self, gc, x, y, angle = 0.0):
        w, h = self.corner_size

        gc.PushState()
        gc.Translate(x, y)
        gc.Rotate(angle)
        path = gc.CreatePath()
        path.MoveToPoint(-w / 2, -h / 2)
        path.AddLineToPoint(w / 2, -h / 2)
        path.AddLineToPoint(w / 2, -h / 2 + h / 4)
        path.AddLineToPoint(w / 12, h / 12)
        path.AddLineToPoint(-w / 2 + w / 4, h / 2)
        path.AddLineToPoint(-w / 2, h / 2)
        path.AddLineToPoint(-w / 2, -h / 2)
        gc.DrawPath(path)
        gc.PopState()

    def highlightCorner(self, gc, corner = 0):
        w, h = self.corner_size
        xinset, yinset = self.corner_inset
        cx, cy = self.center
        ww, wh = self.GetSize()

        if corner == 0:
            x, y = (cx - ww / 2 + xinset + 1, cy - wh / 2 + yinset)
            self.drawCorner(gc, x + w / 2, y + h / 2, 0)
        elif corner == 1:
            x, y = (cx + ww / 2 - xinset, cy - wh / 2 + yinset)
            self.drawCorner(gc, x - w / 2, y + h / 2, math.pi / 2)
        elif corner == 2:
            x, y = (cx + ww / 2 - xinset, cy + wh / 2 - yinset - 1)
            self.drawCorner(gc, x - w / 2, y - h / 2, math.pi)
        elif corner == 3:
            x, y = (cx - ww / 2 + xinset + 1, cy + wh / 2 - yinset - 1)
            self.drawCorner(gc, x + w / 2, y - h / 2, math.pi * 3 / 2)

    def drawCenteredDisc(self, gc, radius):
        cx, cy = self.center
        gc.DrawEllipse(cx - radius, cy - radius, radius * 2, radius * 2)

    def draw(self, dc, w, h):
        dc.SetBackground(wx.Brush(self.bgcolor))
        dc.Clear()
        gc = wx.GraphicsContext.Create(dc)

        if self.bg_bmp:
            w, h = (self.bg_bmp.GetWidth(), self.bg_bmp.GetHeight())
            gc.DrawBitmap(self.bg_bmp, 0, 0, w, h)

        if self.enabled and self.IsEnabled():
            gc.SetPen(wx.Pen(HOVER_PEN_COLOR, 4))
            gc.SetBrush(wx.Brush(HOVER_BRUSH_COLOR))

            if self.concentric is not None:
                if self.concentric < len(self.concentric_circle_radii):
                    if self.concentric == 0:
                        self.drawCenteredDisc(gc, self.concentric_circle_radii[1])
                    elif self.quadrant is not None:
                        self.highlightQuadrant(gc, self.quadrant, self.concentric)
                elif self.corner is not None:
                    self.highlightCorner(gc, self.corner)

            if self.keypad_idx >= 0:
                padw, padh = (self.keypad_bmp.GetWidth(), self.keypad_bmp.GetHeight())
                pos = self.keypad_positions[self.keypad_idx]
                pos = (pos[0] - padw / 2 - 3, pos[1] - padh / 2 - 3)
                gc.DrawBitmap(self.keypad_bmp, pos[0], pos[1], padw, padh)

            if self.hovered_keypad is not None and self.hovered_keypad != self.keypad_idx:
                pos = self.keypad_positions[self.hovered_keypad]
                r = XYButtons.keypad_radius
                gc.DrawEllipse(pos[0]-r/2, pos[1]-r/2, r, r)

            # Draw label overlays
            gc.SetPen(wx.Pen(wx.Colour(255, 255, 255, 128), 1))
            gc.SetBrush(wx.Brush(wx.Colour(255, 255, 255, 128 + 64)))
            for idx, kpos in self.label_overlay_positions.items():
                if idx != self.concentric:
                    r = kpos[2]
                    gc.DrawEllipse(kpos[0] - r, kpos[1] - r, r * 2, r * 2)
        else:
            gc.SetPen(wx.Pen(self.bgcolor, 0))
            gc.SetBrush(wx.Brush(self.bgcolormask))
            gc.DrawRectangle(0, 0, w, h)
    
        self.drawFocusRect(dc)

        # Used to check exact position of keypad dots, should we ever resize the bg image
        # for idx, kpos in self.label_overlay_positions.items():
        #    dc.DrawCircle(kpos[0], kpos[1], kpos[2])

    # ------ #
    # Events #
    # ------ #
    def OnKey(self, evt):
        # print('XYButtons key', evt.GetKeyCode())
        if not self.enabled:
            evt.Skip()
            return
        key = evt.KeyCode
        if self.keypad_idx >= 0:
            if key == wx.WXK_TAB:
                keypad = self.cycleKeypadIndex(not evt.ShiftDown())
                self.setKeypadIndex(keypad)
                if keypad == -1:
                    # exit widget after largest step
                    # evt.Skip()
                    # On MS Windows if tab event is delivered,
                    # it is not handled
                    self.Navigate(not evt.ShiftDown())
                    return
            elif key == wx.WXK_ESCAPE:
                self.setKeypadIndex(-1)
            elif key == wx.WXK_UP:
                self.quadrant = 1
            elif key == wx.WXK_DOWN:
                self.quadrant = 3
            elif key == wx.WXK_LEFT:
                self.quadrant = 2
            elif key == wx.WXK_RIGHT:
                self.quadrant = 0
            elif key == wx.WXK_PAGEUP:
                self.quadrant = 4
            elif key == wx.WXK_PAGEDOWN:
                self.quadrant = 5
            else:
                evt.Skip()
                return

            self.concentric = self.keypad_idx + 1
            
            if self.quadrant is not None:
                x, y, z = self.getMovement(evt)
                if (x or y) and self.moveCallback:
                    self.moveCallback(x, y)
                if z and self.zCallback:
                    self.zCallback(z)
            self.Refresh()
        elif key == wx.WXK_SPACE:
            self.spacebarCallback()
        elif key == wx.WXK_TAB:
            self.setKeypadIndex(len(self.keypad_positions)-1 if evt.ShiftDown() else 0)
        else:
            # handle arrows elsewhere
            evt.Skip()

    def OnMotion(self, event):
        if not self.enabled:
            return

        oldcorner = self.corner
        oldq, oldc = self.quadrant, self.concentric
        old_hovered_keypad = self.hovered_keypad

        mpos = event.GetPosition()
        self.hovered_keypad = self.mouseOverKeypad(mpos)
        self.quadrant = None
        self.concentric = None
        if self.hovered_keypad is None:
            center = wx.Point(self.center[0], self.center[1])
            riseDist = self.distanceToLine(mpos, center.x - 1, center.y - 1, center.x + 1, center.y + 1)
            fallDist = self.distanceToLine(mpos, center.x - 1, center.y + 1, center.x + 1, center.y - 1)
            self.quadrant, self.concentric = self.getQuadrantConcentricFromPosition(mpos)

            # If mouse hovers in space between quadrants, don't commit to a quadrant
            if riseDist <= self.spacer or fallDist <= self.spacer:
                self.quadrant = None

        cx, cy = self.center
        if mpos.x < cx and mpos.y < cy:
            self.corner = 0
        if mpos.x >= cx and mpos.y < cy:
            self.corner = 1
        if mpos.x >= cx and mpos.y >= cy:
            self.corner = 2
        if mpos.x < cx and mpos.y >= cy:
            self.corner = 3

        if oldq != self.quadrant or oldc != self.concentric or oldcorner != self.corner \
            or old_hovered_keypad != self.hovered_keypad:
            self.update()

    def OnLeftDown(self, event):
        if not self.enabled:
            return

        # Take focus when clicked so that arrow keys can control movement
        self.SetFocus()

        mpos = event.GetPosition()

        idx = self.mouseOverKeypad(mpos)
        if idx is None:
            self.quadrant, self.concentric = self.getQuadrantConcentricFromPosition(mpos)
            if self.concentric is not None:
                if self.concentric < len(self.concentric_circle_radii):
                    if self.concentric == 0:
                        self.lastCorner = -1
                        self.lastMove = None
                        self.cornerCallback(self.corner_to_axis[-1])
                    elif self.quadrant is not None:
                        x, y, z = self.getMovement(event)
                        if self.moveCallback:
                            self.lastMove = (x, y)
                            self.lastCorner = None
                            self.moveCallback(x, y)
                elif self.corner is not None:
                    if self.cornerCallback:
                        self.lastCorner = self.corner
                        self.lastMove = None
                        self.cornerCallback(self.corner_to_axis[self.corner])
        else:
            self.setKeypadIndex(-1 if self.keypad_idx == idx else idx)

    def OnLeaveWindow(self, evt):
        self.quadrant = None
        self.concentric = None
        self.update()

class XYButtonsMini(XYButtons):
    imagename = "control_mini.png"
    center = (57, 56.5)
    concentric_circle_radii = [0, 30.3]
    corner_inset = (5, 5)
    corner_size = (50, 50)
    outer_radius = 31
    corner_to_axis = {
        0: "x",
        1: "z",
        2: "y",
        3: "center",
    }

    def bind_events(self):
        # Set up mouse and keyboard event capture
        self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnLeftDown)
        self.Bind(wx.EVT_MOTION, self.OnMotion)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.OnLeaveWindow)

    def OnMotion(self, event):
        if not self.enabled:
            return

        oldcorner = self.corner
        oldq, oldc = self.quadrant, self.concentric

        mpos = event.GetPosition()

        self.quadrant, self.concentric = self.getQuadrantConcentricFromPosition(mpos)

        cx, cy = XYButtonsMini.center
        if mpos.x < cx and mpos.y < cy:
            self.corner = 0
        if mpos.x >= cx and mpos.y < cy:
            self.corner = 1
        if mpos.x >= cx and mpos.y >= cy:
            self.corner = 2
        if mpos.x < cx and mpos.y >= cy:
            self.corner = 3

        if oldq != self.quadrant or oldc != self.concentric or oldcorner != self.corner:
            self.update()

    def OnLeftDown(self, event):
        if not self.enabled:
            return

        # Take focus when clicked so that arrow keys can control movement
        self.SetFocus()

        mpos = event.GetPosition()

        self.quadrant, self.concentric = self.getQuadrantConcentricFromPosition(mpos)
        if self.concentric is not None:
            if self.concentric < len(self.concentric_circle_radii):
                self.cornerCallback("all")
            elif self.corner is not None:
                if self.cornerCallback:
                    self.lastCorner = self.corner
                    self.lastMove = None
                    self.cornerCallback(self.corner_to_axis[self.corner])

    def drawCorner(self, gc, x, y, angle = 0.0):
        w, h = self.corner_size

        gc.PushState()
        gc.Translate(x, y)
        gc.Rotate(angle)
        path = gc.CreatePath()
        path.MoveToPoint(-w / 2, -h / 2)
        path.AddLineToPoint(w / 2, -h / 2)
        path.AddLineToPoint(w / 2, -h / 2 + h / 4)
        path.AddArc(w / 2, h / 2, self.outer_radius, 3 * math.pi / 2, math.pi, False)
        path.AddLineToPoint(-w / 2, h / 2)
        path.AddLineToPoint(-w / 2, -h / 2)
        gc.DrawPath(path)
        gc.PopState()

    def draw(self, dc, w, h):
        dc.SetBackground(wx.Brush(self.bgcolor))
        dc.Clear()
        gc = wx.GraphicsContext.Create(dc)

        if self.bg_bmp:
            w, h = (self.bg_bmp.GetWidth(), self.bg_bmp.GetHeight())
            gc.DrawBitmap(self.bg_bmp, 0, 0, w, h)

        if self.enabled and self.IsEnabled():
            gc.SetPen(wx.Pen(HOVER_PEN_COLOR, 4))
            gc.SetBrush(wx.Brush(HOVER_BRUSH_COLOR))

            if self.concentric is not None:
                if self.concentric < len(self.concentric_circle_radii):
                    self.drawCenteredDisc(gc, self.concentric_circle_radii[-1])
                elif self.corner is not None:
                    self.highlightCorner(gc, self.corner)
        else:
            gc.SetPen(wx.Pen(self.bgcolor, 0))
            gc.SetBrush(wx.Brush(self.bgcolormask))
            gc.DrawRectangle(0, 0, w, h)



#!/usr/bin/env python3

# This file is part of the Printrun suite.
#
# Printrun is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Printrun is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Printrun.  If not, see <http://www.gnu.org/licenses/>.

import wx
import time

import numpy
import pyglet
pyglet.options['debug_gl'] = True

from pyglet.gl import GL_AMBIENT_AND_DIFFUSE, glBegin, glClearColor, \
    glColor3f, GL_CULL_FACE, GL_DEPTH_TEST, GL_DIFFUSE, GL_EMISSION, \
    glEnable, glEnd, GL_FILL, GLfloat, GL_FRONT_AND_BACK, GL_LIGHT0, \
    GL_LIGHT1, glLightfv, GL_LIGHTING, GL_LINE, glMaterialf, glMaterialfv, \
    glMultMatrixd, glNormal3f, glPolygonMode, glPopMatrix, GL_POSITION, \
    glPushMatrix, glRotatef, glScalef, glShadeModel, GL_SHININESS, \
    GL_SMOOTH, GL_SPECULAR, glTranslatef, GL_TRIANGLES, glVertex3f, \
    glGetDoublev, GL_MODELVIEW_MATRIX, GLdouble, glClearDepth, glDepthFunc, \
    GL_LEQUAL, GL_BLEND, glBlendFunc, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, \
    GL_LINE_LOOP, glGetFloatv, GL_LINE_WIDTH, glLineWidth, glDisable, \
    GL_LINE_SMOOTH
from pyglet import gl

from .gl.panel import wxGLPanel
from .gl.trackball import build_rotmatrix
from .gl.libtatlin import actors

def vec(*args):
    return (GLfloat * len(args))(*args)

class stlview:
    def __init__(self, facets, batch):
        # Create the vertex and normal arrays.
        vertices = []
        normals = []

        for i in facets:
            for j in i[1]:
                vertices.extend(j)
                normals.extend(i[0])

        # Create a list of triangle indices.
        indices = list(range(3 * len(facets)))  # [[3*i, 3*i+1, 3*i+2] for i in xrange(len(facets))]
        self.vertex_list = batch.add_indexed(len(vertices) // 3,
                                             GL_TRIANGLES,
                                             None,  # group,
                                             indices,
                                             ('v3f/static', vertices),
                                             ('n3f/static', normals))

    def delete(self):
        self.vertex_list.delete()

class StlViewPanel(wxGLPanel):

    do_lights = False

    def __init__(self, parent, size,
                 build_dimensions = None, circular = False,
                 antialias_samples = 0,
                 grid = (1, 10), perspective=False):
        if perspective:
            self.orthographic=False
        super().__init__(parent, wx.DefaultPosition, size, 0,
                                           antialias_samples = antialias_samples)
        
        self.batches = []
        self.rot = 0
        self.canvas.Bind(wx.EVT_MOUSE_EVENTS, self.move)
        self.canvas.Bind(wx.EVT_MOUSEWHEEL, self.wheel)
        self.canvas.Bind(wx.EVT_LEFT_DCLICK, self.double_click)
        self.initialized = True
        self.parent = parent
        self.initpos = None
        if build_dimensions:
            self.build_dimensions = build_dimensions
        else:
            self.build_dimensions = [200, 200, 100, 0, 0, 0]
        self.platform = actors.Platform(self.build_dimensions,
                                        circular = circular,
                                        grid = grid)
        self.dist = max(self.build_dimensions[0], self.build_dimensions[1])
        self.basequat = [0, 0, 0, 1]
        wx.CallAfter(self.forceresize) #why needed
        self.mousepos = (0, 0)

    def OnReshape(self):
        self.mview_initialized = False
        super(StlViewPanel, self).OnReshape()

    # ==========================================================================
    # GLFrame OpenGL Event Handlers
    # ==========================================================================
    def OnInitGL(self, call_reshape = True):
        '''Initialize OpenGL for use in the window.'''
        if self.GLinitialized:
            return
        self.GLinitialized = True
        # create a pyglet context for this panel
        self.pygletcontext = gl.Context(gl.current_context)
        self.pygletcontext.canvas = self
        self.pygletcontext.set_current()
        # normal gl init
        glClearColor(0, 0, 0, 1)
        glColor3f(1, 0, 0)
        glEnable(GL_DEPTH_TEST)
        glClearDepth(1.0)
        glDepthFunc(GL_LEQUAL)
        glEnable(GL_CULL_FACE)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        # Uncomment this line for a wireframe view
        # glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)

        # Simple light setup.  On Windows GL_LIGHT0 is enabled by default,
        # but this is not the case on Linux or Mac, so remember to always
        # include it.
        glEnable(GL_LIGHTING)
        glEnable(GL_LIGHT0)
        glEnable(GL_LIGHT1)

        glLightfv(GL_LIGHT0, GL_POSITION, vec(.5, .5, 1, 0))
        glLightfv(GL_LIGHT0, GL_SPECULAR, vec(.5, .5, 1, 1))
        glLightfv(GL_LIGHT0, GL_DIFFUSE, vec(1, 1, 1, 1))
        glLightfv(GL_LIGHT1, GL_POSITION, vec(1, 0, .5, 0))
        glLightfv(GL_LIGHT1, GL_DIFFUSE, vec(.5, .5, .5, 1))
        glLightfv(GL_LIGHT1, GL_SPECULAR, vec(1, 1, 1, 1))
        glShadeModel(GL_SMOOTH)

        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, vec(0.5, 0, 0.3, 1))
        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, vec(1, 1, 1, 1))
        glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 50)
        glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, vec(0, 0.1, 0, 0.9))
        if call_reshape:
            self.OnReshape()
        if hasattr(self.parent, "filenames") and self.parent.filenames:
            for filename in self.parent.filenames:
                self.parent.load_file(filename)
            self.parent.autoplate()
            if hasattr(self.parent, "loadcb"):
                self.parent.loadcb()
            self.parent.filenames = None

    def double_click(self, event):
        if hasattr(self.parent, "clickcb") and self.parent.clickcb:
            self.parent.clickcb(event)

    def forceresize(self):
        #print('forceresize')
        x, y = self.GetClientSize()
        #TODO: probably not needed
        self.SetClientSize((x, y+1))
        self.SetClientSize((x, y))
        self.initialized = False

    def move(self, event):
        """react to mouse actions:
        no mouse: show red mousedrop
        LMB: move active object,
            with shift rotate viewport
        RMB: nothing
            with shift move viewport
        """
        self.mousepos = event.GetPosition()
        if event.Dragging():
            if event.LeftIsDown():
                self.handle_rotation(event)
            elif event.RightIsDown():
                self.handle_translation(event)
            self.Refresh(False)
        elif event.ButtonUp(wx.MOUSE_BTN_LEFT) or \
            event.ButtonUp(wx.MOUSE_BTN_RIGHT):
            self.initpos = None
        event.Skip()

    def handle_wheel(self, event):
        delta = event.GetWheelRotation()
        factor = 1.05
        x, y = event.GetPosition()
        x, y, _ = self.mouse_to_3d(x, y, local_transform = True)
        if delta > 0:
            self.zoom(factor, (x, y))
        else:
            self.zoom(1 / factor, (x, y))

    def wheel(self, event):
        """react to mouse wheel actions:
        rotate object
            with shift zoom viewport
        """
        self.handle_wheel(event)
        wx.CallAfter(self.Refresh)

    def keypress(self, event):
        """gets keypress events and moves/rotates active shape"""
        keycode = event.GetKeyCode()
        step = 5
        angle = 18
        if event.ControlDown():
            step = 1
            angle = 1
        # h
        if keycode == 72:
            self.parent.move_shape((-step, 0))
        # l
        if keycode == 76:
            self.parent.move_shape((step, 0))
        # j
        if keycode == 75:
            self.parent.move_shape((0, step))
        # k
        if keycode == 74:
            self.parent.move_shape((0, -step))
        # [
        if keycode == 91:
            self.parent.rotate_shape(-angle)
        # ]
        if keycode == 93:
            self.parent.rotate_shape(angle)
        event.Skip()
        wx.CallAfter(self.Refresh)

    def anim(self, obj):
        g = 50 * 9.8
        v = 20
        dt = 0.05
        basepos = obj.offsets[2]
        obj.offsets[2] += obj.animoffset
        while obj.offsets[2] > -1:
            time.sleep(dt)
            obj.offsets[2] -= v * dt
            v += g * dt
            if obj.offsets[2] < 0:
                obj.scale[2] *= 1 - 3 * dt
        # return
        v = v / 4
        while obj.offsets[2] < basepos:
            time.sleep(dt)
            obj.offsets[2] += v * dt
            v -= g * dt
            obj.scale[2] *= 1 + 5 * dt
        obj.scale[2] = 1.0

    def create_objects(self):
        '''create opengl objects when opengl is initialized'''
        if not self.platform.initialized:
            self.platform.init()
        self.initialized = 1
        #TODO: this probably creates constant redraw
        # create_objects is called during OnDraw, remove
        wx.CallAfter(self.Refresh)

    def prepare_model(self, m, scale):
        batch = pyglet.graphics.Batch()
        stlview(m.facets, batch = batch)
        m.batch = batch
        # m.animoffset = 300
        # threading.Thread(target = self.anim, args = (m, )).start()
        wx.CallAfter(self.Refresh)

    def update_object_resize(self):
        '''called when the window receives only if opengl is initialized'''
        pass

    def draw_objects(self):
        '''called in the middle of ondraw after the buffer has been cleared'''
        self.create_objects()

        glPushMatrix()
        glTranslatef(0, 0, -self.dist)
        glMultMatrixd(build_rotmatrix(self.basequat))  # Rotate according to trackball
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, vec(0.2, 0.2, 0.2, 1))
        glTranslatef(- self.build_dimensions[3] - self.platform.width / 2,
                     - self.build_dimensions[4] - self.platform.depth / 2, 0)  # Move origin to bottom left of platform
        # Draw platform
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        glDisable(GL_LIGHTING)
        self.platform.draw()
        glEnable(GL_LIGHTING)
        # Draw mouse
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        inter = self.mouse_to_plane(self.mousepos[0], self.mousepos[1],
                                    plane_normal = (0, 0, 1), plane_offset = 0,
                                    local_transform = False)
        if inter is not None:
            glPushMatrix()
            glTranslatef(inter[0], inter[1], inter[2])
            glBegin(GL_TRIANGLES)
            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, vec(1, 0, 0, 1))
            glNormal3f(0, 0, 1)
            glVertex3f(2, 2, 0)
            glVertex3f(-2, 2, 0)
            glVertex3f(-2, -2, 0)
            glVertex3f(2, -2, 0)
            glVertex3f(2, 2, 0)
            glVertex3f(-2, -2, 0)
            glEnd()
            glPopMatrix()

        # Draw objects
        glDisable(GL_CULL_FACE)
        glPushMatrix()
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, vec(0.3, 0.7, 0.5, 1))
        for i in self.parent.models:
            model = self.parent.models[i]
            glPushMatrix()
            glTranslatef(*(model.offsets))
            glRotatef(model.rot, 0.0, 0.0, 1.0)
            glTranslatef(*(model.centeroffset))
            glScalef(*model.scale)
            model.batch.draw()
            glPopMatrix()
        glPopMatrix()
        glEnable(GL_CULL_FACE)

        # Draw cutting plane
        if self.parent.cutting:
            # FIXME: make this a proper Actor
            axis = self.parent.cutting_axis
            fixed_dist = self.parent.cutting_dist
            dist, plane_width, plane_height = self.get_cutting_plane(axis, fixed_dist)
            if dist is not None:
                glPushMatrix()
                if axis == "x":
                    glRotatef(90, 0, 1, 0)
                    glRotatef(90, 0, 0, 1)
                    glTranslatef(0, 0, dist)
                elif axis == "y":
                    glRotatef(90, 1, 0, 0)
                    glTranslatef(0, 0, -dist)
                elif axis == "z":
                    glTranslatef(0, 0, dist)
                glDisable(GL_CULL_FACE)
                glBegin(GL_TRIANGLES)
                glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, vec(0, 0.9, 0.15, 0.3))
                glNormal3f(0, 0, self.parent.cutting_direction)
                glVertex3f(plane_width, plane_height, 0)
                glVertex3f(0, plane_height, 0)
                glVertex3f(0, 0, 0)
                glVertex3f(plane_width, 0, 0)
                glVertex3f(plane_width, plane_height, 0)
                glVertex3f(0, 0, 0)
                glEnd()
                glEnable(GL_CULL_FACE)
                glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
                glEnable(GL_LINE_SMOOTH)
                orig_linewidth = (GLfloat)()
                glGetFloatv(GL_LINE_WIDTH, orig_linewidth)
                glLineWidth(4.0)
                glBegin(GL_LINE_LOOP)
                glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, vec(0, 0.8, 0.15, 1))
                glVertex3f(0, 0, 0)
                glVertex3f(0, plane_height, 0)
                glVertex3f(plane_width, plane_height, 0)
                glVertex3f(plane_width, 0, 0)
                glEnd()
                glLineWidth(orig_linewidth)
                glDisable(GL_LINE_SMOOTH)
                glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
                glPopMatrix()

        glPopMatrix()

    # ==========================================================================
    # Utils
    # ==========================================================================
    def get_modelview_mat(self, local_transform):
        mvmat = (GLdouble * 16)()
        if local_transform:
            glPushMatrix()
            # Rotate according to trackball
            glTranslatef(0, 0, -self.dist)
            glMultMatrixd(build_rotmatrix(self.basequat))  # Rotate according to trackball
            glTranslatef(- self.build_dimensions[3] - self.platform.width / 2,
                         - self.build_dimensions[4] - self.platform.depth / 2, 0)  # Move origin to bottom left of platform
            glGetDoublev(GL_MODELVIEW_MATRIX, mvmat)
            glPopMatrix()
        else:
            glGetDoublev(GL_MODELVIEW_MATRIX, mvmat)
        return mvmat

    def get_cutting_plane(self, cutting_axis, fixed_dist, local_transform = False):
        cutting_plane_sizes = {"x": (self.platform.depth, self.platform.height),
                               "y": (self.platform.width, self.platform.height),
                               "z": (self.platform.width, self.platform.depth)}
        plane_width, plane_height = cutting_plane_sizes[cutting_axis]
        if fixed_dist is not None:
            return fixed_dist, plane_width, plane_height
        ref_sizes = {"x": self.platform.width,
                     "y": self.platform.depth,
                     "z": self.platform.height,
                     }
        ref_planes = {"x": (0, 0, 1),
                      "y": (0, 0, 1),
                      "z": (0, 1, 0)
                      }
        ref_offsets = {"x": 0,
                       "y": 0,
                       "z": - self.platform.depth / 2
                       }
        translate_axis = {"x": 0,
                          "y": 1,
                          "z": 2
                          }
        fallback_ref_planes = {"x": (0, 1, 0),
                               "y": (1, 0, 0),
                               "z": (1, 0, 0)
                               }
        fallback_ref_offsets = {"x": - self.platform.height / 2,
                                "y": - self.platform.width / 2,
                                "z": - self.platform.width / 2,
                                }
        ref_size = ref_sizes[cutting_axis]
        ref_plane = ref_planes[cutting_axis]
        ref_offset = ref_offsets[cutting_axis]
        inter = self.mouse_to_plane(self.mousepos[0], self.mousepos[1],
                                    plane_normal = ref_plane,
                                    plane_offset = ref_offset,
                                    local_transform = local_transform)
        max_size = max((self.platform.width,
                        self.platform.depth,
                        self.platform.height))
        dist = None
        if inter is not None and numpy.fabs(inter).max() + max_size / 2 < 2 * max_size:
            dist = inter[translate_axis[cutting_axis]]
        if dist is None or dist < -0.5 * ref_size or dist > 1.5 * ref_size:
            ref_plane = fallback_ref_planes[cutting_axis]
            ref_offset = fallback_ref_offsets[cutting_axis]
            inter = self.mouse_to_plane(self.mousepos[0], self.mousepos[1],
                                        plane_normal = ref_plane,
                                        plane_offset = ref_offset,
                                        local_transform = False)
            if inter is not None and numpy.fabs(inter).max() + max_size / 2 < 2 * max_size:
                dist = inter[translate_axis[cutting_axis]]
        if dist is not None:
            dist = min(1.5 * ref_size, max(-0.5 * ref_size, dist))
        return dist, plane_width, plane_height

def main():
    app = wx.App(redirect = False)
    frame = wx.Frame(None, -1, "GL Window", size = (400, 400))
    StlViewPanel(frame)
    frame.Show(True)
    app.MainLoop()
    app.Destroy()

if __name__ == "__main__":
    main()
